from __future__ import annotations

import importlib
from abc import ABC, abstractmethod
from typing import TYPE_CHECKING, TypeVar

from infrahub.message_bus.messages import ROUTING_KEY_MAP

ResponseClass = TypeVar("ResponseClass")

if TYPE_CHECKING:
    from infrahub.components import ComponentType
    from infrahub.config import BrokerDriver, BrokerSettings
    from infrahub.message_bus import InfrahubMessage, InfrahubResponse
    from infrahub.message_bus.types import MessageTTL
    from infrahub.services import InfrahubServices


class InfrahubMessageBus(ABC):
    DELIVER_TIMEOUT: int = 30 * 60  # 30 minutes
    worker_bindings: list[str] = [
        "check.*.*",
        "event.*.*",
        "finalize.*.*",
        "git.*.*",
        "refresh.webhook.*",
        "request.*.*",
        "send.*.*",
        "schema.*.*",
        "transform.*.*",
        "trigger.*.*",
    ]
    event_bindings: list[str] = ["refresh.registry.*"]
    broadcasted_event_bindings: list[str] = ["refresh.git.*"]
    service: InfrahubServices

    async def shutdown(self) -> None:  # noqa: B027 We want a default empty behavior, so it's ok to have an empty non-abstract method.
        """Shutdown the Message bus"""

    @classmethod
    async def new(cls, component_type: ComponentType, settings: BrokerSettings | None = None) -> InfrahubMessageBus:
        raise NotImplementedError()

    @classmethod
    async def new_from_driver(
        cls, component_type: ComponentType, driver: BrokerDriver, settings: BrokerSettings | None = None
    ) -> InfrahubMessageBus:
        """Imports and initializes the correct class based on the supplied driver.

        This is to ensure that we only import the Python modules that we actually
        need to operate and not import all possible options.
        """
        module = importlib.import_module(driver.driver_module_path)
        broker_driver: InfrahubMessageBus = getattr(module, driver.driver_class_name)
        return await broker_driver.new(component_type=component_type, settings=settings)

    @abstractmethod
    async def publish(
        self, message: InfrahubMessage, routing_key: str, delay: MessageTTL | None = None, is_retry: bool = False
    ) -> None:
        raise NotImplementedError()

    @abstractmethod
    async def reply(self, message: InfrahubMessage, routing_key: str) -> None:
        raise NotImplementedError()

    @abstractmethod
    async def rpc(self, message: InfrahubMessage, response_class: type[ResponseClass]) -> ResponseClass:
        raise NotImplementedError()

    async def send(self, message: InfrahubMessage, delay: MessageTTL | None = None, is_retry: bool = False) -> None:
        routing_key = ROUTING_KEY_MAP.get(type(message))
        if not routing_key:
            raise ValueError("Unable to determine routing key")
        await self.publish(message, routing_key=routing_key, delay=delay, is_retry=is_retry)

    # TODO rename it
    async def reply_if_initiator_meta(self, message: InfrahubResponse, initiator: InfrahubMessage) -> None:
        if initiator.meta:
            message.meta.correlation_id = initiator.meta.correlation_id
            routing_key = initiator.meta.reply_to or ""
            await self.reply(message, routing_key=routing_key)



from __future__ import annotations

from typing import TYPE_CHECKING, Any

from infrahub.core.constants import PathType
from infrahub.core.path import DataPath, GroupedDataPaths

from ..interface import ConstraintCheckerInterface
from ..shared import (
    RelationshipSchemaValidatorQuery,
)

if TYPE_CHECKING:
    from infrahub.core.branch import Branch
    from infrahub.database import InfrahubDatabase

    from ..model import SchemaConstraintValidatorRequest


class RelationshipOptionalUpdateValidatorQuery(RelationshipSchemaValidatorQuery):
    name = "relationship_constraints_optional_validator"

    async def query_init(self, db: InfrahubDatabase, **kwargs: dict[str, Any]) -> None:  # noqa: ARG002
        branch_filter, branch_params = self.branch.get_query_filter_path(at=self.at.to_string(), is_isolated=False)
        self.params.update(branch_params)

        self.params["relationship_id"] = self.relationship_schema.identifier

        query = """
        // Query all Active Nodes of type
        // and store their UUID in uuids_active_node
        MATCH (n:%(node_kind)s)
        CALL {
            WITH n
            MATCH (root:Root)<-[r:IS_PART_OF]-(n)
            WHERE %(branch_filter)s
            RETURN n as n1, r as r1
            ORDER BY r.branch_level DESC, r.from DESC
            LIMIT 1
        }
        WITH n1 as active_node, r1 as rb, n
        WHERE rb.status = "active"
        WITH COLLECT(active_node.uuid) AS uuids_active_node
        // identifier all nodes with at least one active member for this relationship
        // and store their UUID in uuids_with_rel
        MATCH (n:%(node_kind)s)
        CALL {
            WITH n, uuids_active_node
            MATCH path = (n)-[r:IS_RELATED]-(:Relationship { name: $relationship_id })
            WHERE %(branch_filter)s
            RETURN n as n1, r as r1
            ORDER BY r.branch_level DESC, r.from DESC
            LIMIT 1
        }
        WITH n1 as node_with_rel, r1 as r, uuids_active_node
        WHERE r.status = "active"
        WITH COLLECT(node_with_rel.uuid) AS uuids_with_rel, uuids_active_node
        MATCH (n:%(node_kind)s)-[r:IS_PART_OF]->(:Root)
        WHERE n.uuid IN uuids_active_node
          AND not n.uuid IN uuids_with_rel
          AND NOT exists((n)-[:IS_RELATED]-(:Relationship { name: $relationship_id }))
        """ % {"branch_filter": branch_filter, "node_kind": self.node_schema.kind}

        self.add_to_query(query)
        self.return_labels = ["n.uuid", "r as root_relationship"]

    async def get_paths(self) -> GroupedDataPaths:
        grouped_data_paths = GroupedDataPaths()
        for result in self.results:
            grouped_data_paths.add_data_path(
                DataPath(
                    branch=result.get("root_relationship").get("branch"),
                    path_type=PathType.NODE,
                    node_id=str(result.get("n.uuid")),
                    kind=self.node_schema.kind,
                )
            )

        return grouped_data_paths


class RelationshipOptionalChecker(ConstraintCheckerInterface):
    query_classes = [RelationshipOptionalUpdateValidatorQuery]

    def __init__(self, db: InfrahubDatabase, branch: Branch | None = None) -> None:
        self.db = db
        self.branch = branch

    @property
    def name(self) -> str:
        return "relationship.optional.update"

    def supports(self, request: SchemaConstraintValidatorRequest) -> bool:
        return request.constraint_name == self.name

    async def check(self, request: SchemaConstraintValidatorRequest) -> list[GroupedDataPaths]:
        grouped_data_paths_list: list[GroupedDataPaths] = []
        if not request.schema_path.field_name:
            raise ValueError("field_name is not defined")
        relationship_schema = request.node_schema.get_relationship(name=request.schema_path.field_name)
        if relationship_schema.optional is True:
            return grouped_data_paths_list

        for query_class in self.query_classes:
            # TODO add exception handling
            query = await query_class.init(
                db=self.db, branch=self.branch, node_schema=request.node_schema, schema_path=request.schema_path
            )
            await query.execute(db=self.db)
            grouped_data_paths_list.append(await query.get_paths())
        return grouped_data_paths_list



ACCOUNT = "CoreAccount"
ACCOUNTGROUP = "CoreAccountGroup"
ACCOUNTROLE = "CoreAccountRole"
ACCOUNTTOKEN = "InternalAccountToken"
ARTIFACT = "CoreArtifact"
ARTIFACTCHECK = "CoreArtifactCheck"
ARTIFACTDEFINITION = "CoreArtifactDefinition"
ARTIFACTTARGET = "CoreArtifactTarget"
ARTIFACTTHREAD = "CoreArtifactThread"
ARTIFACTVALIDATOR = "CoreArtifactValidator"
BASEPERMISSION = "CoreBasePermission"
CHANGECOMMENT = "CoreChangeComment"
CHANGETHREAD = "CoreChangeThread"
CHECK = "CoreCheck"
CHECKDEFINITION = "CoreCheckDefinition"
COMMENT = "CoreComment"
CUSTOMWEBHOOK = "CoreCustomWebhook"
CREDENTIAL = "CoreCredential"
DATACHECK = "CoreDataCheck"
DATAVALIDATOR = "CoreDataValidator"
FILECHECK = "CoreFileCheck"
FILETHREAD = "CoreFileThread"
GENERICACCOUNT = "CoreGenericAccount"
GENERATORCHECK = "CoreGeneratorCheck"
GENERATORDEFINITION = "CoreGeneratorDefinition"
GENERATORINSTANCE = "CoreGeneratorInstance"
GENERATORVALIDATOR = "CoreGeneratorValidator"
GENERATORGROUP = "CoreGeneratorGroup"
GENERICGROUP = "CoreGroup"
GLOBALPERMISSION = "CoreGlobalPermission"
GRAPHQLQUERY = "CoreGraphQLQuery"
GRAPHQLQUERYGROUP = "CoreGraphQLQueryGroup"
IPNAMESPACE = "BuiltinIPNamespace"
IPADDRESS = "BuiltinIPAddress"
IPADDRESSPOOL = "CoreIPAddressPool"
IPPREFIX = "BuiltinIPPrefix"
IPPREFIXPOOL = "CoreIPPrefixPool"
MENU = "CoreMenu"
MENUITEM = "CoreMenuItem"
NAMESPACE = "IpamNamespace"
NODE = "CoreNode"
NUMBERPOOL = "CoreNumberPool"
LINEAGEOWNER = "LineageOwner"
LINEAGESOURCE = "LineageSource"
OBJECTCOMPONENTTEMPLATE = "CoreObjectComponentTemplate"
OBJECTPERMISSION = "CoreObjectPermission"
OBJECTTEMPLATE = "CoreObjectTemplate"
OBJECTTHREAD = "CoreObjectThread"
PASSWORDCREDENTIAL = "CorePasswordCredential"
PROFILE = "CoreProfile"
PROPOSEDCHANGE = "CoreProposedChange"
REFRESHTOKEN = "InternalRefreshToken"
REPOSITORY = "CoreRepository"
RESOURCEPOOL = "CoreResourcePool"
GENERICREPOSITORY = "CoreGenericRepository"
READONLYREPOSITORY = "CoreReadOnlyRepository"
REPOSITORYVALIDATOR = "CoreRepositoryValidator"
SCHEMACHECK = "CoreSchemaCheck"
SCHEMAVALIDATOR = "CoreSchemaValidator"
STANDARDCHECK = "CoreStandardCheck"
STANDARDGROUP = "CoreStandardGroup"
STANDARDWEBHOOK = "CoreStandardWebhook"
TAG = "BuiltinTag"
TASKTARGET = "CoreTaskTarget"
THREAD = "CoreThread"
THREADCOMMENT = "CoreThreadComment"
TRANSFORM = "CoreTransformation"
TRANSFORMJINJA2 = "CoreTransformJinja2"
TRANSFORMPYTHON = "CoreTransformPython"
USERVALIDATOR = "CoreUserValidator"
VALIDATOR = "CoreValidator"
WEBHOOK = "CoreWebhook"



from abc import ABC, abstractmethod

from infrahub.core.node import Node
from infrahub.core.timestamp import Timestamp


class NodeConstraintInterface(ABC):
    @abstractmethod
    async def check(self, node: Node, at: Timestamp | None = None, filters: list[str] | None = None) -> None: ...



# Generated by "invoke backend.generate", do not edit directly

from __future__ import annotations

from typing import TYPE_CHECKING

from .protocols_base import CoreNode

if TYPE_CHECKING:
    from enum import Enum

    from infrahub.core.attribute import (
        URL,
        Boolean,
        BooleanOptional,
        DateTime,
        DateTimeOptional,
        Dropdown,
        HashedPassword,
        Integer,
        IntegerOptional,
        IPHost,
        IPNetwork,
        JSONAttribute,
        JSONAttributeOptional,
        ListAttributeOptional,
        String,
        StringOptional,
    )
    from infrahub.core.relationship import RelationshipManager


class BuiltinIPAddress(CoreNode):
    address: IPHost
    description: StringOptional
    ip_namespace: RelationshipManager
    ip_prefix: RelationshipManager


class BuiltinIPNamespace(CoreNode):
    name: String
    description: StringOptional
    ip_prefixes: RelationshipManager
    ip_addresses: RelationshipManager


class BuiltinIPPrefix(CoreNode):
    prefix: IPNetwork
    description: StringOptional
    member_type: Dropdown
    is_pool: Boolean
    is_top_level: BooleanOptional
    utilization: IntegerOptional
    netmask: StringOptional
    hostmask: StringOptional
    network_address: StringOptional
    broadcast_address: StringOptional
    ip_namespace: RelationshipManager
    ip_addresses: RelationshipManager
    resource_pool: RelationshipManager
    parent: RelationshipManager
    children: RelationshipManager


class CoreArtifactTarget(CoreNode):
    artifacts: RelationshipManager


class CoreBasePermission(CoreNode):
    description: StringOptional
    identifier: StringOptional
    roles: RelationshipManager


class CoreCheck(CoreNode):
    name: StringOptional
    label: StringOptional
    origin: String
    kind: String
    message: StringOptional
    conclusion: Enum
    severity: Enum
    created_at: DateTimeOptional
    validator: RelationshipManager


class CoreComment(CoreNode):
    text: String
    created_at: DateTimeOptional
    created_by: RelationshipManager


class CoreCredential(CoreNode):
    name: String
    label: StringOptional
    description: StringOptional


class CoreGenericAccount(CoreNode):
    name: String
    password: HashedPassword
    label: StringOptional
    description: StringOptional
    account_type: Enum
    status: Dropdown
    tokens: RelationshipManager


class CoreGenericRepository(CoreNode):
    name: String
    description: StringOptional
    location: String
    internal_status: Dropdown
    operational_status: Dropdown
    sync_status: Dropdown
    credential: RelationshipManager
    tags: RelationshipManager
    transformations: RelationshipManager
    queries: RelationshipManager
    checks: RelationshipManager
    generators: RelationshipManager


class CoreGroup(CoreNode):
    name: String
    label: StringOptional
    description: StringOptional
    group_type: Enum
    members: RelationshipManager
    subscribers: RelationshipManager
    parent: RelationshipManager
    children: RelationshipManager


class CoreMenu(CoreNode):
    namespace: String
    name: String
    label: StringOptional
    kind: StringOptional
    path: StringOptional
    description: StringOptional
    icon: StringOptional
    protected: Boolean
    order_weight: Integer
    required_permissions: ListAttributeOptional
    section: Enum
    parent: RelationshipManager
    children: RelationshipManager


class CoreObjectComponentTemplate(CoreNode):
    template_name: String


class CoreObjectTemplate(CoreNode):
    template_name: String


class CoreProfile(CoreNode):
    profile_name: String
    profile_priority: IntegerOptional


class CoreResourcePool(CoreNode):
    name: String
    description: StringOptional


class CoreTaskTarget(CoreNode):
    pass


class CoreThread(CoreNode):
    label: StringOptional
    resolved: Boolean
    created_at: DateTimeOptional
    change: RelationshipManager
    comments: RelationshipManager
    created_by: RelationshipManager


class CoreTransformation(CoreNode):
    name: String
    label: StringOptional
    description: StringOptional
    timeout: Integer
    query: RelationshipManager
    repository: RelationshipManager
    tags: RelationshipManager


class CoreValidator(CoreNode):
    label: StringOptional
    state: Enum
    conclusion: Enum
    completed_at: DateTimeOptional
    started_at: DateTimeOptional
    proposed_change: RelationshipManager
    checks: RelationshipManager


class CoreWebhook(CoreNode):
    name: String
    event_type: Enum
    branch_scope: Dropdown
    node_kind: StringOptional
    description: StringOptional
    url: URL
    validate_certificates: BooleanOptional


class LineageOwner(CoreNode):
    pass


class LineageSource(CoreNode):
    pass


class BuiltinTag(CoreNode):
    name: String
    description: StringOptional


class CoreAccount(LineageOwner, LineageSource, CoreGenericAccount):
    pass


class CoreAccountGroup(LineageOwner, LineageSource, CoreGroup):
    roles: RelationshipManager


class CoreAccountRole(CoreNode):
    name: String
    groups: RelationshipManager
    permissions: RelationshipManager


class CoreArtifact(CoreTaskTarget):
    name: String
    status: Enum
    content_type: Enum
    checksum: StringOptional
    storage_id: StringOptional
    parameters: JSONAttributeOptional
    object: RelationshipManager
    definition: RelationshipManager


class CoreArtifactCheck(CoreCheck):
    changed: BooleanOptional
    checksum: StringOptional
    artifact_id: StringOptional
    storage_id: StringOptional
    line_number: IntegerOptional


class CoreArtifactDefinition(CoreTaskTarget):
    name: String
    artifact_name: String
    description: StringOptional
    parameters: JSONAttribute
    content_type: Enum
    targets: RelationshipManager
    transformation: RelationshipManager


class CoreArtifactThread(CoreThread):
    artifact_id: StringOptional
    storage_id: StringOptional
    line_number: IntegerOptional


class CoreArtifactValidator(CoreValidator):
    definition: RelationshipManager


class CoreChangeComment(CoreComment):
    change: RelationshipManager


class CoreChangeThread(CoreThread):
    pass


class CoreCheckDefinition(CoreTaskTarget):
    name: String
    description: StringOptional
    file_path: String
    class_name: String
    timeout: Integer
    parameters: JSONAttributeOptional
    repository: RelationshipManager
    query: RelationshipManager
    targets: RelationshipManager
    tags: RelationshipManager


class CoreCustomWebhook(CoreWebhook, CoreTaskTarget):
    transformation: RelationshipManager


class CoreDataCheck(CoreCheck):
    conflicts: JSONAttribute
    keep_branch: Enum
    enriched_conflict_id: StringOptional


class CoreDataValidator(CoreValidator):
    pass


class CoreFileCheck(CoreCheck):
    files: ListAttributeOptional
    commit: StringOptional


class CoreFileThread(CoreThread):
    file: StringOptional
    commit: StringOptional
    line_number: IntegerOptional
    repository: RelationshipManager


class CoreGeneratorCheck(CoreCheck):
    instance: String


class CoreGeneratorDefinition(CoreTaskTarget):
    name: String
    description: StringOptional
    parameters: JSONAttribute
    file_path: String
    class_name: String
    convert_query_response: BooleanOptional
    query: RelationshipManager
    repository: RelationshipManager
    targets: RelationshipManager


class CoreGeneratorGroup(CoreGroup):
    pass


class CoreGeneratorInstance(CoreTaskTarget):
    name: String
    status: Enum
    object: RelationshipManager
    definition: RelationshipManager


class CoreGeneratorValidator(CoreValidator):
    definition: RelationshipManager


class CoreGlobalPermission(CoreBasePermission):
    action: Dropdown
    decision: Enum


class CoreGraphQLQuery(CoreNode):
    name: String
    description: StringOptional
    query: String
    variables: JSONAttributeOptional
    operations: ListAttributeOptional
    models: ListAttributeOptional
    depth: IntegerOptional
    height: IntegerOptional
    repository: RelationshipManager
    tags: RelationshipManager


class CoreGraphQLQueryGroup(CoreGroup):
    parameters: JSONAttributeOptional
    query: RelationshipManager


class CoreIPAddressPool(CoreResourcePool, LineageSource):
    default_address_type: String
    default_prefix_length: IntegerOptional
    resources: RelationshipManager
    ip_namespace: RelationshipManager


class CoreIPPrefixPool(CoreResourcePool, LineageSource):
    default_prefix_length: IntegerOptional
    default_member_type: Enum
    default_prefix_type: StringOptional
    resources: RelationshipManager
    ip_namespace: RelationshipManager


class CoreMenuItem(CoreMenu):
    pass


class CoreNumberPool(CoreResourcePool, LineageSource):
    node: String
    node_attribute: String
    start_range: Integer
    end_range: Integer


class CoreObjectPermission(CoreBasePermission):
    namespace: String
    name: String
    action: Enum
    decision: Enum


class CoreObjectThread(CoreThread):
    object_path: String


class CorePasswordCredential(CoreCredential):
    username: StringOptional
    password: StringOptional


class CoreProposedChange(CoreTaskTarget):
    name: String
    description: StringOptional
    source_branch: String
    destination_branch: String
    state: Enum
    approved_by: RelationshipManager
    reviewers: RelationshipManager
    created_by: RelationshipManager
    comments: RelationshipManager
    threads: RelationshipManager
    validations: RelationshipManager


class CoreReadOnlyRepository(LineageOwner, LineageSource, CoreGenericRepository, CoreTaskTarget):
    ref: String
    commit: StringOptional


class CoreRepository(LineageOwner, LineageSource, CoreGenericRepository, CoreTaskTarget):
    default_branch: String
    commit: StringOptional


class CoreRepositoryValidator(CoreValidator):
    repository: RelationshipManager


class CoreSchemaCheck(CoreCheck):
    conflicts: JSONAttribute
    enriched_conflict_id: StringOptional


class CoreSchemaValidator(CoreValidator):
    pass


class CoreStandardCheck(CoreCheck):
    pass


class CoreStandardGroup(CoreGroup):
    pass


class CoreStandardWebhook(CoreWebhook, CoreTaskTarget):
    shared_key: String


class CoreThreadComment(CoreComment):
    thread: RelationshipManager


class CoreTransformJinja2(CoreTransformation):
    template_path: String


class CoreTransformPython(CoreTransformation):
    file_path: String
    class_name: String


class CoreUserValidator(CoreValidator):
    check_definition: RelationshipManager
    repository: RelationshipManager


class InternalAccountToken(CoreNode):
    name: StringOptional
    token: String
    expiration: DateTimeOptional
    account: RelationshipManager


class InternalRefreshToken(CoreNode):
    expiration: DateTime
    account: RelationshipManager


class IpamNamespace(BuiltinIPNamespace):
    default: BooleanOptional



from infrahub.core.constants import RelationshipKind, SchemaPathType
from infrahub.core.constants.schema import UpdateSupport
from infrahub.core.diff.model.path import NodeDiffFieldSummary
from infrahub.core.models import SchemaUpdateConstraintInfo
from infrahub.core.path import SchemaPath
from infrahub.core.schema import AttributeSchema, MainSchemaTypes
from infrahub.core.schema.relationship_schema import RelationshipSchema
from infrahub.core.schema.schema_branch import SchemaBranch
from infrahub.core.validators import CONSTRAINT_VALIDATOR_MAP
from infrahub.log import get_logger

LOG = get_logger(__name__)


class ConstraintValidatorDeterminer:
    def __init__(self, schema_branch: SchemaBranch) -> None:
        self.schema_branch = schema_branch
        self._node_kinds: set[str] = set()
        self._attribute_element_map: dict[str, set[str]] = {}
        self._relationship_element_map: dict[str, set[str]] = {}

    def _index_node_diffs(self, node_diffs: list[NodeDiffFieldSummary]) -> None:
        for node_diff in node_diffs:
            self._node_kinds.add(node_diff.kind)
            if node_diff.kind not in self._attribute_element_map:
                self._attribute_element_map[node_diff.kind] = set()
            for attribute_name in node_diff.attribute_names:
                self._attribute_element_map[node_diff.kind].add(attribute_name)
            if node_diff.kind not in self._relationship_element_map:
                self._relationship_element_map[node_diff.kind] = set()
            for relationship_name in node_diff.relationship_names:
                self._relationship_element_map[node_diff.kind].add(relationship_name)

    def _has_attribute_diff(self, kind: str, name: str) -> bool:
        return name in self._attribute_element_map.get(kind, set())

    def _has_relationship_diff(self, kind: str, name: str) -> bool:
        return name in self._relationship_element_map.get(kind, set())

    async def get_constraints(
        self, node_diffs: list[NodeDiffFieldSummary], filter_invalid: bool = True
    ) -> list[SchemaUpdateConstraintInfo]:
        self._index_node_diffs(node_diffs)
        constraints: list[SchemaUpdateConstraintInfo] = []
        if not node_diffs:
            return constraints

        constraints.extend(await self._get_all_property_constraints())

        for kind in self._node_kinds:
            schema = self.schema_branch.get(name=kind, duplicate=False)
            constraints.extend(await self._get_constraints_for_one_schema(schema))

        if not filter_invalid:
            return constraints

        validated_constraints: list[SchemaUpdateConstraintInfo] = []
        for constraint in constraints:
            if CONSTRAINT_VALIDATOR_MAP.get(constraint.constraint_name, None):
                validated_constraints.append(constraint)
            else:
                LOG.warning(
                    f"Unable to validate: {constraint.constraint_name!r} for {constraint.path.get_path()!r}, validator not available",
                    constraint_name=constraint.constraint_name,
                    path=constraint.path.get_path(),
                )

        return validated_constraints

    async def _get_constraints_for_one_schema(self, schema: MainSchemaTypes) -> list[SchemaUpdateConstraintInfo]:
        constraints: list[SchemaUpdateConstraintInfo] = []
        constraints.extend(await self._get_attribute_constraints_for_one_schema(schema=schema))
        constraints.extend(await self._get_relationship_constraints_for_one_schema(schema=schema))
        return constraints

    async def _get_all_property_constraints(self) -> list[SchemaUpdateConstraintInfo]:
        constraints: list[SchemaUpdateConstraintInfo] = []
        for schema in self.schema_branch.get_all().values():
            constraints.extend(await self._get_property_constraints_for_one_schema(schema=schema))
        return constraints

    async def _get_property_constraints_for_one_schema(
        self, schema: MainSchemaTypes
    ) -> list[SchemaUpdateConstraintInfo]:
        constraints: list[SchemaUpdateConstraintInfo] = []
        for prop_name, prop_field_info in schema.model_fields.items():
            if (
                prop_name in ["attributes", "relationships"]
                or not prop_field_info.json_schema_extra
                or not isinstance(prop_field_info.json_schema_extra, dict)
            ):
                continue

            prop_field_update = prop_field_info.json_schema_extra.get("update")
            if prop_field_update != UpdateSupport.VALIDATE_CONSTRAINT.value:
                continue

            if getattr(schema, prop_name) is None:
                continue

            schema_path = SchemaPath(
                schema_kind=schema.kind,
                path_type=SchemaPathType.NODE,
                field_name=prop_name,
                property_name=prop_name,
            )
            constraint_name = f"node.{prop_name}.update"

            constraints.append(SchemaUpdateConstraintInfo(constraint_name=constraint_name, path=schema_path))
        return constraints

    async def _get_attribute_constraints_for_one_schema(
        self, schema: MainSchemaTypes
    ) -> list[SchemaUpdateConstraintInfo]:
        constraints: list[SchemaUpdateConstraintInfo] = []
        for field_name in schema.attribute_names:
            if self._has_attribute_diff(kind=schema.kind, name=field_name):
                field = schema.get_attribute(field_name)
                constraints.extend(await self._get_constraints_for_one_field(schema=schema, field=field))
        return constraints

    async def _get_relationship_constraints_for_one_schema(
        self, schema: MainSchemaTypes
    ) -> list[SchemaUpdateConstraintInfo]:
        constraints: list[SchemaUpdateConstraintInfo] = []
        for field_name in schema.relationship_names:
            if self._has_relationship_diff(kind=schema.kind, name=field_name):
                field = schema.get_relationship(field_name)
                constraints.extend(await self._get_constraints_for_one_field(schema=schema, field=field))
        return constraints

    async def _get_constraints_for_one_field(
        self, schema: MainSchemaTypes, field: AttributeSchema | RelationshipSchema
    ) -> list[SchemaUpdateConstraintInfo]:
        constraints: list[SchemaUpdateConstraintInfo] = []
        for prop_name, prop_field_info in field.model_fields.items():
            if not prop_field_info.json_schema_extra or not isinstance(prop_field_info.json_schema_extra, dict):
                continue

            prop_field_update = prop_field_info.json_schema_extra.get("update")
            if prop_field_update != UpdateSupport.VALIDATE_CONSTRAINT.value:
                continue

            if getattr(field, prop_name) is None:
                continue

            path_type = SchemaPathType.ATTRIBUTE
            constraint_name = f"attribute.{prop_name}.update"
            if isinstance(field, RelationshipSchema):
                if field.kind == RelationshipKind.GROUP:
                    continue
                path_type = SchemaPathType.RELATIONSHIP
                constraint_name = f"relationship.{prop_name}.update"

            schema_path = SchemaPath(
                schema_kind=schema.kind,
                path_type=path_type,
                field_name=field.name,
                property_name=prop_name,
            )

            constraints.append(SchemaUpdateConstraintInfo(constraint_name=constraint_name, path=schema_path))
        return constraints



from __future__ import annotations

from pathlib import Path

from dulwich import porcelain
from dulwich.repo import Repo


class GitRepoManager:
    def __init__(self, root_directory: str, branch: str = "main"):
        self.root_directory = root_directory
        self.branch = branch
        self.git: Repo = self.initialize_repo()

    def initialize_repo(self) -> Repo:
        # Check if the directory already has a repository

        root_path = Path(self.root_directory)

        if root_path.exists() and (root_path / ".git").is_dir():
            repo = Repo(self.root_directory)  # Open existing repo
        else:
            repo = Repo.init(self.root_directory, default_branch=self.branch.encode("utf-8"))

        if not repo:
            raise ValueError("Failed to initialize or open a repository.")

        return repo

    @property
    def active_branch(self) -> str | None:
        active_branch = porcelain.active_branch(self.root_directory).decode("utf-8")
        return active_branch



from typing import Any

from ...generic_schema import GenericSchema
from ...node_schema import NodeSchema
from .account import (
    core_account,
    core_account_token,
    core_credential,
    core_generic_account,
    core_password_credential,
    core_refresh_token,
)
from .artifact import core_artifact, core_artifact_definition, core_artifact_target
from .builtin import builtin_tag
from .check import core_check_definition
from .core import core_node, core_task_target
from .generator import core_generator_definition, core_generator_instance
from .graphql_query import core_graphql_query
from .group import core_generator_group, core_graphql_query_group, core_group, core_standard_group
from .ipam import builtin_ip_address, builtin_ip_prefix, builtin_ipam, core_ipam_namespace
from .lineage import lineage_owner, lineage_source
from .menu import generic_menu_item, menu_item
from .permission import (
    core_account_group,
    core_account_role,
    core_base_permission,
    core_global_permission,
    core_object_permission,
)
from .profile import core_profile_schema_definition
from .propose_change import core_proposed_change
from .propose_change_comment import (
    core_artifact_thread,
    core_change_comment,
    core_change_thread,
    core_file_thread,
    core_object_thread,
    core_propose_change_comment,
    core_thread,
    core_thread_comment,
)
from .propose_change_validator import (
    core_artifact_check,
    core_artifact_validator,
    core_check,
    core_data_check,
    core_data_validator,
    core_file_check,
    core_generator_check,
    core_generator_validator,
    core_propose_change_validator,
    core_repository_validator,
    core_schema_check,
    core_schema_validator,
    core_standard_check,
    core_user_validator,
)
from .repository import core_generic_repository, core_read_only_repository, core_repository
from .resource_pool import core_ip_address_pool, core_ip_prefix_pool, core_number_pool, core_resource_pool
from .template import core_object_component_template, core_object_template
from .transform import core_transform, core_transform_jinja2, core_transform_python
from .webhook import core_custom_webhook, core_standard_webhook, core_webhook

core_models_mixed: dict[str, list] = {
    "generics": [
        core_node,
        lineage_owner,
        core_profile_schema_definition,
        lineage_source,
        core_propose_change_comment,
        core_thread,
        core_group,
        core_propose_change_validator,
        core_check,
        core_transform,
        core_artifact_target,
        core_task_target,
        core_webhook,
        core_generic_repository,
        builtin_ipam,
        builtin_ip_prefix,
        builtin_ip_address,
        core_resource_pool,
        core_generic_account,
        core_base_permission,
        core_credential,
        core_object_template,
        core_object_component_template,
        generic_menu_item,
    ],
    "nodes": [
        menu_item,
        core_standard_group,
        core_generator_group,
        core_graphql_query_group,
        builtin_tag,
        core_account,
        core_account_token,
        core_password_credential,
        core_refresh_token,
        core_proposed_change,
        core_change_thread,
        core_file_thread,
        core_artifact_thread,
        core_object_thread,
        core_change_comment,
        core_thread_comment,
        core_repository,
        core_read_only_repository,
        core_transform_jinja2,
        core_data_check,
        core_standard_check,
        core_schema_check,
        core_file_check,
        core_artifact_check,
        core_generator_check,
        core_data_validator,
        core_repository_validator,
        core_user_validator,
        core_schema_validator,
        core_artifact_validator,
        core_generator_validator,
        core_check_definition,
        core_transform_python,
        core_graphql_query,
        core_artifact,
        core_artifact_definition,
        core_generator_definition,
        core_generator_instance,
        core_standard_webhook,
        core_custom_webhook,
        core_ipam_namespace,
        core_ip_prefix_pool,
        core_ip_address_pool,
        core_number_pool,
        core_global_permission,
        core_object_permission,
        core_account_role,
        core_account_group,
    ],
}


core_models: dict[str, Any] = {
    "generics": [item.to_dict() if isinstance(item, GenericSchema) else item for item in core_models_mixed["generics"]],
    "nodes": [item.to_dict() if isinstance(item, NodeSchema) else item for item in core_models_mixed["nodes"]],
}



from __future__ import annotations

from typing import TYPE_CHECKING, Any

from graphene import Argument, Boolean, DateTime, Field, InputObjectType, Int, List, NonNull, ObjectType, String
from graphene import Enum as GrapheneEnum
from infrahub_sdk.utils import extract_fields

from infrahub.core import registry
from infrahub.core.constants import DiffAction, RelationshipCardinality
from infrahub.core.constants.database import DatabaseEdgeType
from infrahub.core.diff.model.path import NameTrackingId
from infrahub.core.diff.repository.repository import DiffRepository
from infrahub.core.query.diff import DiffCountChanges
from infrahub.core.timestamp import Timestamp
from infrahub.dependencies.registry import get_component_registry
from infrahub.graphql.enums import ConflictSelection as GraphQLConflictSelection

if TYPE_CHECKING:
    from datetime import datetime

    from graphql import GraphQLResolveInfo

    from infrahub.core.diff.model.path import (
        EnrichedDiffAttribute,
        EnrichedDiffConflict,
        EnrichedDiffNode,
        EnrichedDiffProperty,
        EnrichedDiffRelationship,
        EnrichedDiffRoot,
        EnrichedDiffSingleRelationship,
    )
    from infrahub.database import InfrahubDatabase
    from infrahub.graphql.initialization import GraphqlContext

GrapheneDiffActionEnum = GrapheneEnum.from_enum(DiffAction)
GrapheneCardinalityEnum = GrapheneEnum.from_enum(RelationshipCardinality)


class ConflictDetails(ObjectType):
    uuid = String(required=True)
    base_branch_action = Field(GrapheneDiffActionEnum, required=True)
    base_branch_value = String()
    base_branch_changed_at = DateTime(required=True)
    base_branch_label = String()
    diff_branch_action = Field(GrapheneDiffActionEnum, required=True)
    diff_branch_value = String()
    diff_branch_changed_at = DateTime(required=True)
    diff_branch_label = String()
    selected_branch = Field(GraphQLConflictSelection)
    resolvable = Boolean()


class DiffSummaryCounts(ObjectType):
    num_added = Int(required=False)
    num_updated = Int(required=False)
    num_removed = Int(required=False)
    num_conflicts = Int(required=False)


class DiffProperty(ObjectType):
    property_type = String(required=True)
    last_changed_at = DateTime(required=True)
    previous_value = String(required=False)
    new_value = String(required=False)
    previous_label = String(required=False)
    new_label = String(required=False)
    status = Field(GrapheneDiffActionEnum, required=True)
    path_identifier = String(required=True)
    conflict = Field(ConflictDetails, required=False)


class DiffAttribute(DiffSummaryCounts):
    name = String(required=True)
    last_changed_at = DateTime(required=True)
    status = Field(GrapheneDiffActionEnum, required=True)
    path_identifier = String(required=True)
    properties = List(NonNull(DiffProperty))
    contains_conflict = Boolean(required=True)
    conflict = Field(ConflictDetails, required=False)


class DiffSingleRelationship(DiffSummaryCounts):
    last_changed_at = DateTime(required=False)
    status = Field(GrapheneDiffActionEnum, required=True)
    peer_id = String(required=True)
    peer_label = String(required=False)
    path_identifier = String(required=True)
    contains_conflict = Boolean(required=True)
    conflict = Field(ConflictDetails, required=False)
    properties = List(NonNull(DiffProperty))


class DiffRelationship(DiffSummaryCounts):
    name = String(required=True)
    label = String(required=False)
    last_changed_at = DateTime(required=False)
    cardinality = Field(GrapheneCardinalityEnum, required=True)
    status = Field(GrapheneDiffActionEnum, required=True)
    path_identifier = String(required=True)
    elements = List(NonNull(DiffSingleRelationship), required=True)
    contains_conflict = Boolean(required=True)


class DiffNodeParent(ObjectType):
    uuid = String(required=True)
    kind = String(required=False)
    relationship_name = String(required=False)


class DiffNode(DiffSummaryCounts):
    uuid = String(required=True)
    kind = String(required=True)
    label = String(required=True)
    status = Field(GrapheneDiffActionEnum, required=True)
    path_identifier = String(required=True)
    conflict = Field(ConflictDetails, required=False)
    contains_conflict = Boolean(required=True)
    last_changed_at = DateTime(required=False)
    parent = Field(DiffNodeParent, required=False)
    attributes = List(NonNull(DiffAttribute), required=True)
    relationships = List(NonNull(DiffRelationship), required=True)


class DiffTree(DiffSummaryCounts):
    base_branch = String(required=True)
    diff_branch = String(required=True)
    from_time = DateTime(required=True)
    to_time = DateTime(required=True)
    num_untracked_base_changes = Int(required=False)
    num_untracked_diff_changes = Int(required=False)
    name = String(required=False)
    nodes = List(NonNull(DiffNode))


class DiffTreeSummary(DiffSummaryCounts):
    base_branch = String(required=True)
    diff_branch = String(required=True)
    from_time = DateTime(required=True)
    to_time = DateTime(required=True)
    num_unchanged = Int(required=False)
    num_untracked_base_changes = Int(required=False)
    num_untracked_diff_changes = Int(required=False)


class DiffTreeResolver:
    async def to_diff_tree(
        self, enriched_diff_root: EnrichedDiffRoot, graphql_context: GraphqlContext | None = None
    ) -> DiffTree:
        all_nodes = list(enriched_diff_root.nodes)
        tree_nodes = [self.to_diff_node(enriched_node=e_node, graphql_context=graphql_context) for e_node in all_nodes]
        name = None
        if enriched_diff_root.tracking_id and isinstance(enriched_diff_root.tracking_id, NameTrackingId):
            name = enriched_diff_root.tracking_id.name
        return DiffTree(
            base_branch=enriched_diff_root.base_branch_name,
            diff_branch=enriched_diff_root.diff_branch_name,
            from_time=await enriched_diff_root.from_time.to_graphql(),
            to_time=await enriched_diff_root.to_time.to_graphql(),
            name=name,
            nodes=tree_nodes,
            num_added=enriched_diff_root.num_added,
            num_updated=enriched_diff_root.num_updated,
            num_removed=enriched_diff_root.num_removed,
            num_conflicts=enriched_diff_root.num_conflicts,
        )

    def to_diff_node(self, enriched_node: EnrichedDiffNode, graphql_context: GraphqlContext | None = None) -> DiffNode:
        diff_attributes = [
            self.to_diff_attribute(enriched_attribute=e_attr, graphql_context=graphql_context)
            for e_attr in enriched_node.attributes
        ]
        diff_relationships = [
            self.to_diff_relationship(enriched_relationship=e_rel, graphql_context=graphql_context)
            for e_rel in enriched_node.relationships
            if e_rel.include_in_response
        ]
        conflict = None
        if enriched_node.conflict:
            conflict = self.to_diff_conflict(enriched_conflict=enriched_node.conflict, graphql_context=graphql_context)

        parent = None
        if parent_info := enriched_node.get_parent_info(graphql_context=graphql_context):
            parent = DiffNodeParent(
                uuid=parent_info.node.uuid,
                kind=parent_info.node.kind,
                relationship_name=parent_info.relationship_name,
            )

        return DiffNode(
            uuid=enriched_node.uuid,
            kind=enriched_node.kind,
            label=enriched_node.label,
            status=enriched_node.action,
            parent=parent,
            last_changed_at=enriched_node.changed_at.to_datetime() if enriched_node.changed_at else None,
            path_identifier=enriched_node.path_identifier,
            attributes=diff_attributes,
            relationships=diff_relationships,
            contains_conflict=enriched_node.contains_conflict,
            conflict=conflict,
            num_added=enriched_node.num_added,
            num_updated=enriched_node.num_updated,
            num_removed=enriched_node.num_removed,
            num_conflicts=enriched_node.num_conflicts,
        )

    def to_diff_attribute(
        self, enriched_attribute: EnrichedDiffAttribute, graphql_context: GraphqlContext | None = None
    ) -> DiffAttribute:
        diff_properties = [
            self.to_diff_property(enriched_property=e_prop, graphql_context=graphql_context)
            for e_prop in enriched_attribute.properties
        ]
        conflict = None
        for diff_prop in diff_properties:
            if diff_prop.property_type == DatabaseEdgeType.HAS_VALUE.value and diff_prop.conflict:
                conflict = diff_prop.conflict
                diff_prop.conflict = None
        return DiffAttribute(
            name=enriched_attribute.name,
            last_changed_at=enriched_attribute.changed_at.to_datetime(),
            status=enriched_attribute.action,
            path_identifier=enriched_attribute.path_identifier,
            properties=diff_properties,
            contains_conflict=enriched_attribute.contains_conflict,
            conflict=conflict,
            num_added=enriched_attribute.num_added,
            num_updated=enriched_attribute.num_updated,
            num_removed=enriched_attribute.num_removed,
            num_conflicts=enriched_attribute.num_conflicts,
        )

    def to_diff_relationship(
        self, enriched_relationship: EnrichedDiffRelationship, graphql_context: GraphqlContext | None = None
    ) -> DiffRelationship:
        diff_elements = [
            self.to_diff_relationship_element(enriched_element=element, graphql_context=graphql_context)
            for element in enriched_relationship.relationships
        ]
        return DiffRelationship(
            name=enriched_relationship.name,
            label=enriched_relationship.label,
            last_changed_at=enriched_relationship.changed_at.to_datetime()
            if enriched_relationship.changed_at
            else None,
            status=enriched_relationship.action,
            cardinality=enriched_relationship.cardinality,
            path_identifier=enriched_relationship.path_identifier,
            elements=diff_elements,
            contains_conflict=enriched_relationship.contains_conflict,
            num_added=enriched_relationship.num_added,
            num_updated=enriched_relationship.num_updated,
            num_removed=enriched_relationship.num_removed,
            num_conflicts=enriched_relationship.num_conflicts,
        )

    def to_diff_relationship_element(
        self, enriched_element: EnrichedDiffSingleRelationship, graphql_context: GraphqlContext | None = None
    ) -> DiffSingleRelationship:
        diff_properties = [self.to_diff_property(e_prop) for e_prop in enriched_element.properties]
        conflict = None
        if enriched_element.conflict:
            conflict = self.to_diff_conflict(
                enriched_conflict=enriched_element.conflict, graphql_context=graphql_context
            )
        return DiffSingleRelationship(
            last_changed_at=enriched_element.changed_at.to_datetime(),
            status=enriched_element.action,
            peer_id=enriched_element.peer_id,
            peer_label=enriched_element.peer_label,
            path_identifier=enriched_element.path_identifier,
            conflict=conflict,
            properties=diff_properties,
            contains_conflict=enriched_element.contains_conflict,
            num_added=enriched_element.num_added,
            num_updated=enriched_element.num_updated,
            num_removed=enriched_element.num_removed,
            num_conflicts=enriched_element.num_conflicts,
        )

    def to_diff_property(
        self, enriched_property: EnrichedDiffProperty, graphql_context: GraphqlContext | None = None
    ) -> DiffProperty:
        conflict = None
        if enriched_property.conflict:
            conflict = self.to_diff_conflict(
                enriched_conflict=enriched_property.conflict, graphql_context=graphql_context
            )
        return DiffProperty(
            property_type=enriched_property.property_type.value,
            last_changed_at=enriched_property.changed_at.to_datetime(),
            previous_value=enriched_property.previous_value,
            new_value=enriched_property.new_value,
            previous_label=enriched_property.previous_label,
            new_label=enriched_property.new_label,
            status=enriched_property.action,
            path_identifier=enriched_property.path_identifier,
            conflict=conflict,
        )

    def to_diff_conflict(
        self,
        enriched_conflict: EnrichedDiffConflict,
        graphql_context: GraphqlContext | None = None,  # noqa: ARG002
    ) -> ConflictDetails:
        return ConflictDetails(
            uuid=enriched_conflict.uuid,
            base_branch_action=enriched_conflict.base_branch_action,
            base_branch_value=enriched_conflict.base_branch_value,
            base_branch_changed_at=enriched_conflict.base_branch_changed_at.to_datetime()
            if enriched_conflict.base_branch_changed_at
            else None,
            base_branch_label=enriched_conflict.base_branch_label,
            diff_branch_action=enriched_conflict.diff_branch_action,
            diff_branch_value=enriched_conflict.diff_branch_value,
            diff_branch_changed_at=enriched_conflict.diff_branch_changed_at.to_datetime()
            if enriched_conflict.diff_branch_changed_at
            else None,
            diff_branch_label=enriched_conflict.diff_branch_label,
            selected_branch=enriched_conflict.selected_branch.value if enriched_conflict.selected_branch else None,
            resolvable=enriched_conflict.resolvable,
        )

    async def to_graphql(
        self, fields: dict[str, dict], diff_object: Any | None
    ) -> list[dict[str, Any]] | dict[str, Any] | None:
        if diff_object is None:
            return None
        if isinstance(diff_object, list):
            list_response = True
            diff_elements = diff_object
        else:
            list_response = False
            diff_elements = [diff_object]

        response_list = []
        for diff_object_element in diff_elements:
            element_response = {}
            for field_name, sub_fields in fields.items():
                if sub_fields is None:
                    element_response[field_name] = getattr(diff_object_element, field_name, None)
                elif hasattr(diff_object_element, field_name):
                    element_response[field_name] = await self.to_graphql(
                        sub_fields, getattr(diff_object_element, field_name)
                    )
                else:
                    continue
            response_list.append(element_response)
        if list_response:
            return response_list
        return response_list[0]

    async def _add_untracked_fields(
        self,
        db: InfrahubDatabase,
        diff_response: DiffTreeSummary | DiffTree,
        from_time: Timestamp,
        base_branch_name: str | None = None,
        diff_branch_name: str | None = None,
    ) -> None:
        if not (base_branch_name or diff_branch_name):
            return
        branch_names = []
        if base_branch_name:
            branch_names.append(base_branch_name)
        if diff_branch_name:
            branch_names.append(diff_branch_name)
        query = await DiffCountChanges.init(db=db, branch_names=branch_names, diff_from=from_time, diff_to=Timestamp())
        await query.execute(db=db)
        branch_change_map = query.get_num_changes_by_branch()
        if base_branch_name:
            diff_response.num_untracked_base_changes = branch_change_map.get(base_branch_name, 0)
        if diff_branch_name:
            diff_response.num_untracked_diff_changes = branch_change_map.get(diff_branch_name, 0)

    async def resolve(
        self,
        root: dict,  # noqa: ARG002
        info: GraphQLResolveInfo,
        branch: str | None = None,
        name: str | None = None,
        from_time: datetime | None = None,
        to_time: datetime | None = None,
        filters: dict | None = None,
        include_parents: bool = True,
        root_node_uuids: list[str] | None = None,
        limit: int | None = None,
        offset: int | None = None,
    ) -> list[dict[str, Any]] | dict[str, Any] | None:
        component_registry = get_component_registry()
        graphql_context: GraphqlContext = info.context
        base_branch = await registry.get_branch(db=graphql_context.db, branch=registry.default_branch)
        diff_branch = await registry.get_branch(db=graphql_context.db, branch=branch)
        diff_repo = await component_registry.get_component(DiffRepository, db=graphql_context.db, branch=diff_branch)
        branch_start_timestamp = Timestamp(diff_branch.get_branched_from())
        if from_time:
            from_timestamp = Timestamp(from_time.isoformat())
        else:
            from_timestamp = branch_start_timestamp
        if to_time:
            to_timestamp = Timestamp(to_time.isoformat())
        else:
            to_timestamp = graphql_context.at or Timestamp()

        # Convert filters to dict and merge root_node_uuids for compatibility
        filters_dict = dict(filters or {})
        if root_node_uuids and "ids" in filters_dict and isinstance(filters_dict["ids"], list):
            filters_dict["ids"].extend(root_node_uuids)
        elif root_node_uuids:
            filters_dict["ids"] = root_node_uuids

        enriched_diffs = await diff_repo.get(
            base_branch_name=base_branch.name,
            diff_branch_names=[diff_branch.name],
            from_time=from_timestamp,
            to_time=to_timestamp,
            filters=filters_dict,
            include_parents=include_parents,
            limit=limit,
            offset=offset,
            tracking_id=NameTrackingId(name) if name else None,
            include_empty=True,
        )
        if not enriched_diffs:
            return None
        if len(enriched_diffs) > 0:
            # take the one with the longest duration that covers multiple branches
            enriched_diff = sorted(
                enriched_diffs,
                key=lambda d: (
                    d.base_branch_name != d.diff_branch_name,
                    d.to_time.to_datetime() - d.from_time.to_datetime(),
                ),
                reverse=True,
            )[0]
        else:
            enriched_diff = enriched_diffs[0]

        full_fields = await extract_fields(info.field_nodes[0].selection_set)
        diff_tree = await self.to_diff_tree(enriched_diff_root=enriched_diff, graphql_context=graphql_context)
        need_base_changes = "num_untracked_base_changes" in full_fields
        need_branch_changes = "num_untracked_diff_changes" in full_fields
        if need_base_changes or need_branch_changes:
            await self._add_untracked_fields(
                db=graphql_context.db,
                diff_response=diff_tree,
                from_time=enriched_diff.to_time,
                base_branch_name=base_branch.name if need_base_changes else None,
                diff_branch_name=diff_branch.name if need_branch_changes else None,
            )
        return await self.to_graphql(fields=full_fields, diff_object=diff_tree)

    async def summary(
        self,
        root: dict,  # noqa: ARG002
        info: GraphQLResolveInfo,
        branch: str | None = None,
        from_time: datetime | None = None,
        to_time: datetime | None = None,
        filters: dict | None = None,
    ) -> list[dict[str, Any]] | dict[str, Any] | None:
        component_registry = get_component_registry()
        graphql_context: GraphqlContext = info.context
        base_branch = await registry.get_branch(db=graphql_context.db, branch=registry.default_branch)
        diff_branch = await registry.get_branch(db=graphql_context.db, branch=branch)
        diff_repo = await component_registry.get_component(DiffRepository, db=graphql_context.db, branch=diff_branch)
        branch_start_timestamp = Timestamp(diff_branch.get_branched_from())
        if from_time:
            from_timestamp = Timestamp(from_time.isoformat())
        else:
            from_timestamp = branch_start_timestamp
        if to_time:
            to_timestamp = Timestamp(to_time.isoformat())
        else:
            to_timestamp = graphql_context.at or Timestamp()

        filters_dict = dict(filters or {})

        summary = await diff_repo.summary(
            base_branch_name=base_branch.name,
            diff_branch_names=[diff_branch.name],
            from_time=from_timestamp,
            to_time=to_timestamp,
            filters=filters_dict,
        )
        if summary is None:
            return None

        diff_tree_summary = DiffTreeSummary(
            base_branch=base_branch.name,
            diff_branch=diff_branch.name,
            from_time=summary.from_time.to_datetime(),
            to_time=summary.to_time.to_datetime(),
            **summary.model_dump(exclude={"from_time", "to_time"}),
        )
        full_fields = await extract_fields(info.field_nodes[0].selection_set)
        need_base_changes = "num_untracked_base_changes" in full_fields
        need_branch_changes = "num_untracked_diff_changes" in full_fields
        if need_base_changes or need_branch_changes:
            await self._add_untracked_fields(
                db=graphql_context.db,
                diff_response=diff_tree_summary,
                from_time=summary.to_time,
                base_branch_name=base_branch.name if need_base_changes else None,
                diff_branch_name=diff_branch.name if need_branch_changes else None,
            )
        return diff_tree_summary


class IncExclFilterOptions(InputObjectType):
    includes = List(String)
    excludes = List(String)


class IncExclFilterStatusOptions(InputObjectType):
    includes = List(GrapheneDiffActionEnum)
    excludes = List(GrapheneDiffActionEnum)


class DiffTreeQueryFilters(InputObjectType):
    ids = List(String)
    status = IncExclFilterStatusOptions()
    kind = IncExclFilterOptions()
    namespace = IncExclFilterOptions()


DiffTreeQuery = Field(
    DiffTree,
    name=String(),
    branch=String(),
    from_time=DateTime(),
    to_time=DateTime(),
    root_node_uuids=Argument(List(String), deprecation_reason="replaced by filters"),
    include_parents=Boolean(),
    filters=DiffTreeQueryFilters(),
    limit=Int(),
    offset=Int(),
    resolver=DiffTreeResolver().resolve,
    required=False,
)

DiffTreeSummaryQuery = Field(
    DiffTreeSummary,
    name=String(),
    branch=String(),
    from_time=DateTime(),
    to_time=DateTime(),
    filters=DiffTreeQueryFilters(),
    resolver=DiffTreeResolver().summary,
    required=False,
)



from __future__ import annotations

import ipaddress
from typing import TYPE_CHECKING, Any

from graphene import Boolean, Field, Int, List, NonNull, ObjectType, String
from infrahub_sdk.utils import extract_fields_first_node, is_valid_uuid

from infrahub.core.constants import InfrahubKind
from infrahub.core.manager import NodeManager

if TYPE_CHECKING:
    from graphql import GraphQLResolveInfo

    from infrahub.core.protocols import CoreNode
    from infrahub.graphql.initialization import GraphqlContext


class Node(ObjectType):
    id = Field(String, required=True)
    kind = Field(String, required=True, description="The node kind")


class NodeEdge(ObjectType):
    node = Field(Node, required=True)


class NodeEdges(ObjectType):
    count = Field(Int, required=True)
    edges = Field(List(of_type=NonNull(NodeEdge)), required=True)


def _collapse_ipv6(s: str) -> str:
    """Collapse an ipv6 address, ipv6 network, or a partial ipv6 address in extended format, into its collapsed form.
    Raises an error if input does not resemble an IPv6 address in extended format. It means this function also raises
    an error if input string is the start of an IPv6 address in collapsed format.
    """

    try:
        return str(ipaddress.IPv6Address(s))
    except ipaddress.AddressValueError:
        pass

    try:
        return ipaddress.IPv6Network(s).with_prefixlen
    except ipaddress.AddressValueError:
        pass

    # Input string might be an incomplete address in IPv6 format,
    # in which case we would like the collapsed form equivalent of this incomplete address for matching purposes.
    # To get it, we first try to pad the incomplete address with zeros, then we retrieve the collapsed form
    # of the full address, and we remove extra "::" or ":0" at the end of it.

    error_message = "Input string does not match IPv6 extended format"

    # Input string cannot be an IPv6 in extended format if it contains ":"
    if "::" in s:
        raise ValueError(error_message)

    # Add padding to complete the address if needed
    segments = s.split(":")

    if len(segments) == 0:
        raise ValueError(error_message)

    # If any of the non-last segments has less than 4 characters it means we deal with
    # a IPv6 collapsed form or an invalid address
    for segment in segments[:-1]:
        if len(segment) != 4:
            raise ValueError(error_message)

    # Add 0 padding to last segment
    if len(segments[-1]) > 4:
        raise ValueError(error_message)

    segments[-1] += "0" * (4 - len(segments[-1]))

    # Complete the address to have 8 segments by padding with zeros
    while len(segments) < 8:
        segments.append("0000")

    # Create a full IPv6 address from the partial input
    full_address = ":".join(segments)

    # Create an IPv6Address object for validation and to build IPv6 collapsed form.
    ipv6_address = ipaddress.IPv6Address(full_address)

    compressed_address = ipv6_address.compressed

    # We padded with zeros so address might endswith "::" or ":0".
    if compressed_address.endswith(("::", ":0")):
        return compressed_address[:-2]

    # Otherwise, it means 8th segment of ipv6 address was not full and not composed of 0 only
    # e.g. 2001:0db8:0000:0000:0000:0000:03
    return compressed_address


async def search_resolver(
    root: dict,  # noqa: ARG001
    info: GraphQLResolveInfo,
    q: str,
    limit: int = 10,
    partial_match: bool = True,
) -> dict[str, Any]:
    graphql_context: GraphqlContext = info.context
    response: dict[str, Any] = {}
    results: list[CoreNode] = []

    fields = await extract_fields_first_node(info)

    if is_valid_uuid(q):
        matching: CoreNode | None = await NodeManager.get_one(
            db=graphql_context.db, branch=graphql_context.branch, at=graphql_context.at, id=q
        )
        if matching:
            results.append(matching)
    else:
        try:
            # Convert any IPv6 address, network or partial address to collapsed format as it might be stored in db.
            q = _collapse_ipv6(q)
        except (ValueError, ipaddress.AddressValueError):
            pass

        for kind in [InfrahubKind.NODE, InfrahubKind.GENERICGROUP]:
            objs = await NodeManager.query(
                db=graphql_context.db,
                branch=graphql_context.branch,
                schema=kind,
                filters={"any__value": q},
                limit=limit,
                partial_match=partial_match,
            )
            results.extend(objs)

    if "edges" in fields:
        response["edges"] = [{"node": {"id": obj.id, "kind": obj.get_kind()}} for obj in results]

    if "count" in fields:
        response["count"] = len(results)

    return response


InfrahubSearchAnywhere = Field(
    NodeEdges,
    q=String(required=True),
    limit=Int(required=False),
    partial_match=Boolean(required=False),
    resolver=search_resolver,
    required=True,
)



from __future__ import annotations

from typing import TYPE_CHECKING, Any

from infrahub.core.query import Query, QueryType

if TYPE_CHECKING:
    from infrahub.core.branch import Branch
    from infrahub.core.timestamp import Timestamp
    from infrahub.database import InfrahubDatabase


class DiffMergeQuery(Query):
    name = "diff_merge"
    type = QueryType.WRITE
    insert_return = False

    def __init__(
        self,
        node_diff_dicts: dict[str, Any],
        at: Timestamp,
        target_branch: Branch,
        **kwargs: Any,
    ) -> None:
        super().__init__(**kwargs)
        self.node_diff_dicts = node_diff_dicts
        self.at = at
        self.target_branch = target_branch
        self.source_branch_name = self.branch.name

    async def query_init(self, db: InfrahubDatabase, **kwargs: Any) -> None:  # noqa: ARG002
        self.params = {
            "node_diff_dicts": self.node_diff_dicts,
            "at": self.at.to_string(),
            "branch_level": self.target_branch.hierarchy_level,
            "target_branch": self.target_branch.name,
            "source_branch": self.source_branch_name,
        }
        # ruff: noqa: E501
        query = """
UNWIND $node_diff_dicts AS node_diff_map
CALL {
    WITH node_diff_map
    MATCH (n:Node {uuid: node_diff_map.uuid})
    RETURN n
}
WITH n, node_diff_map
CALL {
    WITH n, node_diff_map
    WITH n, node_diff_map, CASE
        WHEN node_diff_map.action = "ADDED" THEN "active"
        WHEN node_diff_map.action = "REMOVED" THEN "deleted"
        ELSE NULL
    END AS node_rel_status
    CALL {
        // ------------------------------
        // only make IS_PART_OF updates if node is ADDED or REMOVED
        // ------------------------------
        WITH n, node_diff_map, node_rel_status
        WITH n, node_diff_map, node_rel_status
        WHERE node_rel_status IS NOT NULL
        MATCH (root:Root)
        // ------------------------------
        // set IS_PART_OF.to, optionally, target branch
        // ------------------------------
        WITH root, n, node_rel_status
        CALL {
            WITH root, n, node_rel_status
            OPTIONAL MATCH (root)<-[target_r_root:IS_PART_OF {branch: $target_branch, status: "active"}]-(n)
            WHERE node_rel_status = "deleted"
            AND target_r_root.from <= $at AND target_r_root.to IS NULL
            SET target_r_root.to = $at
        }
        // ------------------------------
        // create new IS_PART_OF relationship on target_branch
        // ------------------------------
        WITH root, n, node_rel_status
        CALL {
            WITH root, n, node_rel_status
            OPTIONAL MATCH (root)<-[r_root:IS_PART_OF {branch: $target_branch}]-(n)
            WHERE r_root.status = node_rel_status
            AND r_root.from <= $at
            AND (r_root.to >= $at OR r_root.to IS NULL)
            WITH root, r_root, n, node_rel_status
            WHERE r_root IS NULL
            CREATE (root)
                <-[:IS_PART_OF { branch: $target_branch, branch_level: $branch_level, from: $at, status: node_rel_status }]
                -(n)
        }
        // ------------------------------
        // shortcut to delete all attributes and relationships for this node if the node is deleted
        // ------------------------------
        CALL {
            WITH n, node_rel_status
            WITH n, node_rel_status
            WHERE node_rel_status = "deleted"
            CALL {
                WITH n
                OPTIONAL MATCH (n)-[rel1:IS_RELATED]-(:Relationship)-[rel2]-(p)
                WHERE (p.uuid IS NULL OR n.uuid <> p.uuid)
                AND rel1.branch = $target_branch
                AND rel2.branch = $target_branch
                AND rel1.status = "active"
                AND rel2.status = "active"
                RETURN rel1, rel2
                UNION
                WITH n
                OPTIONAL MATCH (n)-[rel1:HAS_ATTRIBUTE]->(:Attribute)-[rel2]->()
                WHERE type(rel2) <> "HAS_ATTRIBUTE"
                AND rel1.branch = $target_branch
                AND rel2.branch = $target_branch
                AND rel1.status = "active"
                AND rel2.status = "active"
                RETURN rel1, rel2
            }
            WITH n, rel1, rel2
            WHERE rel1.to IS NULL
            AND rel2.to IS NULL
            AND rel1.from <= $at
            AND rel2.from <= $at
            SET rel1.to = $at
            SET rel2.to = $at
            // ------------------------------
            // and delete HAS_OWNER and HAS_SOURCE edges to this node if the node is deleted
            // ------------------------------
            WITH n
            CALL {
                WITH n
                CALL {
                    WITH n
                    MATCH (n)<-[rel:HAS_OWNER]-()
                    WHERE rel.branch = $target_branch
                    AND rel.status = "active"
                    AND rel.from <= $at
                    AND rel.to IS NULL
                    RETURN rel
                    UNION
                    MATCH (n)<-[rel:HAS_SOURCE]-()
                    WHERE rel.branch = $target_branch
                    AND rel.status = "active"
                    AND rel.from <= $at
                    AND rel.to IS NULL
                    RETURN rel
                }
                SET rel.to = $at
            }
        }
    }
    WITH n, node_diff_map
    CALL {
        WITH n, node_diff_map
        WITH n, CASE
            WHEN node_diff_map.attributes IS NULL OR node_diff_map.attributes = [] THEN [NULL]
            ELSE node_diff_map.attributes
        END AS attribute_maps
        UNWIND attribute_maps AS attribute_diff_map
        // ------------------------------
        // handle updates for attributes under this node
        // ------------------------------
        CALL {
            WITH n, attribute_diff_map
            WITH n, attribute_diff_map.name AS attr_name, CASE
                WHEN attribute_diff_map.action = "ADDED" THEN "active"
                WHEN attribute_diff_map.action = "REMOVED" THEN "deleted"
                ELSE NULL
            END AS attr_rel_status
            CALL {
                WITH n, attr_name
                OPTIONAL MATCH (n)-[has_attr:HAS_ATTRIBUTE]->(a:Attribute {name: attr_name})
                WHERE has_attr.branch IN [$source_branch, $target_branch]
                RETURN a
                ORDER BY has_attr.from DESC
                LIMIT 1
            }
            WITH n, attr_rel_status, a
            // ------------------------------
            // set HAS_ATTRIBUTE.to on target branch if necessary
            // ------------------------------
            CALL {
                WITH n, attr_rel_status, a
                OPTIONAL MATCH (n)
                    -[target_r_attr:HAS_ATTRIBUTE {branch: $target_branch, status: "active"}]
                    ->(a)
                WHERE attr_rel_status = "deleted"
                AND target_r_attr.from <= $at AND target_r_attr.to IS NULL
                SET target_r_attr.to = $at
            }
            WITH n, attr_rel_status, a
            // ------------------------------
            // conditionally create new HAS_ATTRIBUTE relationship on target_branch, if necessary
            // ------------------------------
            CALL {
                WITH n, attr_rel_status, a
                WITH n, attr_rel_status, a
                WHERE a IS NOT NULL
                OPTIONAL MATCH (n)-[r_attr:HAS_ATTRIBUTE {branch: $target_branch}]->(a)
                WHERE r_attr.status = attr_rel_status
                AND r_attr.from <= $at
                AND (r_attr.to >= $at OR r_attr.to IS NULL)
                WITH n, r_attr, attr_rel_status, a
                WHERE r_attr IS NULL
                CREATE (n)-[:HAS_ATTRIBUTE { branch: $target_branch, branch_level: $branch_level, from: $at, status: attr_rel_status }]->(a)
            }
            RETURN 1 AS done
        }
        RETURN 1 AS done
    }
    WITH n, node_diff_map
    CALL {
        WITH n,node_diff_map
        UNWIND node_diff_map.relationships AS relationship_diff_map
        // ------------------------------
        // handle updates for relationships under this node
        // ------------------------------
        CALL {
            WITH n, relationship_diff_map
            WITH n, relationship_diff_map.peer_id AS rel_peer_id, relationship_diff_map.name AS rel_name, CASE
                WHEN relationship_diff_map.action = "ADDED" THEN "active"
                WHEN relationship_diff_map.action = "REMOVED" THEN "deleted"
                ELSE NULL
            END AS related_rel_status
            // ------------------------------
            // determine the directions of each IS_RELATED
            // ------------------------------
            CALL {
                WITH n, rel_name, rel_peer_id, related_rel_status
                MATCH (n)
                    -[source_r_rel_1:IS_RELATED]
                    -(r:Relationship {name: rel_name})
                    -[source_r_rel_2:IS_RELATED]
                    -(:Node {uuid: rel_peer_id})
                WHERE source_r_rel_1.branch IN [$source_branch, $target_branch]
                AND source_r_rel_2.branch IN [$source_branch, $target_branch]
                AND source_r_rel_1.from <= $at AND source_r_rel_1.to IS NULL
                AND source_r_rel_2.from <= $at AND source_r_rel_2.to IS NULL
                WITH n, rel_name, rel_peer_id, related_rel_status, r, source_r_rel_1, source_r_rel_2
                ORDER BY source_r_rel_1.branch_level DESC, source_r_rel_2.branch_level DESC, source_r_rel_1.from DESC, source_r_rel_2.from DESC
                LIMIT 1
                RETURN r, CASE
                    WHEN startNode(source_r_rel_1).uuid = n.uuid THEN "r"
                    ELSE "l"
                END AS r1_dir,
                CASE
                    WHEN startNode(source_r_rel_2).uuid = r.uuid THEN "r"
                    ELSE "l"
                END AS r2_dir,
                source_r_rel_1.hierarchy AS r1_hierarchy,
                source_r_rel_2.hierarchy AS r2_hierarchy
            }
            WITH n, r, r1_dir, r2_dir, r1_hierarchy, r2_hierarchy, rel_name, rel_peer_id, related_rel_status
            CALL {
                WITH n, rel_name, rel_peer_id, related_rel_status
                OPTIONAL MATCH (n)
                    -[target_r_rel_1:IS_RELATED {branch: $target_branch, status: "active"}]
                    -(:Relationship {name: rel_name})
                    -[target_r_rel_2:IS_RELATED {branch: $target_branch, status: "active"}]
                    -(:Node {uuid: rel_peer_id})
                WHERE related_rel_status = "deleted"
                AND target_r_rel_1.from <= $at AND target_r_rel_1.to IS NULL
                AND target_r_rel_2.from <= $at AND target_r_rel_2.to IS NULL
                SET target_r_rel_1.to = $at
                SET target_r_rel_2.to = $at
            }
            WITH n, r, r1_dir, r2_dir, r1_hierarchy, r2_hierarchy, rel_name, rel_peer_id, related_rel_status
            // ------------------------------
            // conditionally create new IS_RELATED relationships on target_branch, if necessary
            // ------------------------------
            CALL {
                WITH n, r, r1_dir, r2_dir, r1_hierarchy, r2_hierarchy, rel_name, rel_peer_id, related_rel_status
                MATCH (p:Node {uuid: rel_peer_id})
                OPTIONAL MATCH (n)
                    -[r_rel_1:IS_RELATED {branch: $target_branch, status: related_rel_status}]
                    -(:Relationship {name: rel_name})
                    -[r_rel_2:IS_RELATED {branch: $target_branch, status: related_rel_status}]
                    -(p)
                WHERE r_rel_1.from <= $at
                AND (r_rel_1.to >= $at OR r_rel_1.to IS NULL)
                AND r_rel_2.from <= $at
                AND (r_rel_2.to >= $at OR r_rel_2.to IS NULL)
                WITH n, r, r1_dir, r2_dir, r1_hierarchy, r2_hierarchy, p, related_rel_status, r_rel_1, r_rel_2
                WHERE r_rel_1 IS NULL
                AND r_rel_2 IS NULL
                // ------------------------------
                // create IS_RELATED relationships with directions maintained from source
                // ------------------------------
                CALL {
                    WITH n, r, r1_dir, r1_hierarchy, related_rel_status
                    WITH n, r, r1_dir, r1_hierarchy, related_rel_status
                    WHERE r1_dir = "r"
                    CREATE (n)
                        -[:IS_RELATED {branch: $target_branch, branch_level: $branch_level, from: $at, status: related_rel_status, hierarchy: r1_hierarchy}]
                        ->(r)
                }
                CALL {
                    WITH n, r, r1_dir, r1_hierarchy, related_rel_status
                    WITH n, r, r1_dir, r1_hierarchy, related_rel_status
                    WHERE r1_dir = "l"
                    CREATE (n)
                        <-[:IS_RELATED {branch: $target_branch, branch_level: $branch_level, from: $at, status: related_rel_status, hierarchy: r1_hierarchy}]
                        -(r)
                }
                CALL {
                    WITH r, p, r2_dir, r2_hierarchy, related_rel_status
                    WITH r, p, r2_dir, r2_hierarchy, related_rel_status
                    WHERE r2_dir = "r"
                    CREATE (r)
                        -[:IS_RELATED {branch: $target_branch, branch_level: $branch_level, from: $at, status: related_rel_status, hierarchy: r2_hierarchy}]
                        ->(p)
                }
                CALL {
                    WITH r, p, r2_dir, r2_hierarchy, related_rel_status
                    WITH r, p, r2_dir, r2_hierarchy, related_rel_status
                    WHERE r2_dir = "l"
                    CREATE (r)
                        <-[:IS_RELATED {branch: $target_branch, branch_level: $branch_level, from: $at, status: related_rel_status, hierarchy: r2_hierarchy}]
                        -(p)
                }
            }
        }
    }
}
RETURN 1 AS done
        """
        self.add_to_query(query=query)


class DiffMergePropertiesQuery(Query):
    name = "diff_merge_properties"
    type = QueryType.WRITE
    insert_return = False

    def __init__(
        self,
        property_diff_dicts: dict[str, Any],
        at: Timestamp,
        target_branch: Branch,
        **kwargs: Any,
    ) -> None:
        super().__init__(**kwargs)
        self.property_diff_dicts = property_diff_dicts
        self.at = at
        self.target_branch = target_branch
        self.source_branch_name = self.branch.name

    async def query_init(self, db: InfrahubDatabase, **kwargs: Any) -> None:  # noqa: ARG002
        self.params = {
            "property_diff_dicts": self.property_diff_dicts,
            "at": self.at.to_string(),
            "branch_level": self.target_branch.hierarchy_level,
            "target_branch": self.target_branch.name,
            "source_branch": self.source_branch_name,
        }
        query = """
UNWIND $property_diff_dicts AS attr_rel_prop_diff
CALL {
    // ------------------------------
    // find the Attribute node
    // ------------------------------
    WITH attr_rel_prop_diff
    CALL {
        WITH attr_rel_prop_diff
        OPTIONAL MATCH (n:Node {uuid: attr_rel_prop_diff.node_uuid})
            -[has_attr:HAS_ATTRIBUTE]
            ->(attr:Attribute {name: attr_rel_prop_diff.attribute_name})
        WHERE attr_rel_prop_diff.attribute_name IS NOT NULL
        AND has_attr.branch IN [$source_branch, $target_branch]
        RETURN attr
        ORDER BY has_attr.from DESC
        LIMIT 1
    }
    CALL {
        WITH attr_rel_prop_diff
        OPTIONAL MATCH (n:Node {uuid: attr_rel_prop_diff.node_uuid})
            -[r1:IS_RELATED]
            -(rel:Relationship {name: attr_rel_prop_diff.relationship_id})
            -[r2:IS_RELATED]
            -(:Node {uuid: attr_rel_prop_diff.peer_uuid})
        WHERE attr_rel_prop_diff.relationship_id IS NOT NULL
        AND r1.branch IN [$source_branch, $target_branch]
        AND r2.branch IN [$source_branch, $target_branch]
        RETURN rel
        ORDER BY r1.branch_level DESC, r2.branch_level DESC, r1.from DESC, r2.from DESC
        LIMIT 1
    }
    WITH attr_rel_prop_diff, COALESCE(attr, rel) AS attr_rel
    UNWIND attr_rel_prop_diff.properties AS property_diff
    // ------------------------------
    // handle updates for properties under this attribute/relationship
    // ------------------------------
    CALL {
        WITH attr_rel, property_diff
        // ------------------------------
        // identify the correct property node to link
        // ------------------------------
        CALL {
            WITH attr_rel, property_diff
            OPTIONAL MATCH (peer:Node {uuid: property_diff.value})
            WHERE property_diff.property_type IN ["HAS_SOURCE", "HAS_OWNER"]
            // ------------------------------
            // the serialized diff might not include the values for IS_VISIBLE and IS_PROTECTED in
            // some cases, so we need to figure them out here
            // ------------------------------
            CALL {
                WITH attr_rel, property_diff
                OPTIONAL MATCH (attr_rel)-[r_vis_pro]->(bool:Boolean)
                WHERE property_diff.property_type IN ["IS_VISIBLE", "IS_PROTECTED"]
                AND r_vis_pro.branch IN [$source_branch, $target_branch]
                AND type(r_vis_pro) = property_diff.property_type
                AND (property_diff.value IS NULL OR bool.value = property_diff.value)
                RETURN bool
                ORDER BY r_vis_pro.from DESC
                LIMIT 1
            }
            CALL {
                // ------------------------------
                // get the latest linked AttributeValue on the source b/c there could be multiple
                // with different is_default values
                // ------------------------------
                WITH attr_rel, property_diff
                OPTIONAL MATCH (attr_rel)-[r_attr_val:HAS_VALUE]->(av:AttributeValue)
                WHERE property_diff.property_type = "HAS_VALUE"
                AND (
                    av.value = property_diff.value
                    OR toLower(toString(av.value)) = toLower(toString(property_diff.value))
                )
                AND r_attr_val.branch IN [$source_branch, $target_branch]
                RETURN av
                ORDER BY r_attr_val.from DESC
                LIMIT 1
            }
            RETURN COALESCE (peer, bool, av) AS prop_node
        }
        WITH attr_rel, property_diff.property_type AS prop_type, prop_node, CASE
            WHEN property_diff.action = "ADDED" THEN "active"
            WHEN property_diff.action = "REMOVED" THEN "deleted"
            ELSE NULL
        END as prop_rel_status
        // ------------------------------
        // set property edge.to, optionally, on target branch
        // ------------------------------
        CALL {
            WITH attr_rel, prop_rel_status, prop_type
            OPTIONAL MATCH (attr_rel)
                -[target_r_prop {branch: $target_branch}]
                ->()
            WHERE type(target_r_prop) = prop_type
            AND target_r_prop.from < $at AND target_r_prop.to IS NULL
            SET target_r_prop.to = $at
        }
        // ------------------------------
        // check for existing edge on target_branch
        // ------------------------------
        CALL {
            WITH attr_rel, prop_rel_status, prop_type, prop_node
            OPTIONAL MATCH (attr_rel)-[r_prop {branch: $target_branch}]->(prop_node)
            WHERE type(r_prop) = prop_type
            AND r_prop.status = prop_rel_status
            AND r_prop.from <= $at
            AND (r_prop.to > $at OR r_prop.to IS NULL)
            RETURN r_prop
        }
        WITH attr_rel, prop_rel_status, prop_type, prop_node, r_prop
        WHERE r_prop IS NULL
        // ------------------------------
        // create new edge to prop_node on target_branch, if necessary
        // one subquery per possible edge type b/c edge type cannot be a variable
        // ------------------------------
        CALL {
            WITH attr_rel, prop_rel_status, prop_type, prop_node
            WITH attr_rel, prop_rel_status, prop_type, prop_node
            WHERE prop_type = "HAS_VALUE"
            CREATE (attr_rel)-[:HAS_VALUE { branch: $target_branch, branch_level: $branch_level, from: $at, status: prop_rel_status }]->(prop_node)
        }
        CALL {
            WITH attr_rel, prop_rel_status, prop_type, prop_node
            WITH attr_rel, prop_rel_status, prop_type, prop_node
            WHERE prop_type = "HAS_SOURCE"
            CREATE (attr_rel)-[:HAS_SOURCE { branch: $target_branch, branch_level: $branch_level, from: $at, status: prop_rel_status }]->(prop_node)
        }
        CALL {
            WITH attr_rel, prop_rel_status, prop_type, prop_node
            WITH attr_rel, prop_rel_status, prop_type, prop_node
            WHERE prop_type = "HAS_OWNER"
            CREATE (attr_rel)-[:HAS_OWNER { branch: $target_branch, branch_level: $branch_level, from: $at, status: prop_rel_status }]->(prop_node)
        }
        CALL {
            WITH attr_rel, prop_rel_status, prop_type, prop_node
            WITH attr_rel, prop_rel_status, prop_type, prop_node
            WHERE prop_type = "IS_VISIBLE"
            CREATE (attr_rel)-[:IS_VISIBLE { branch: $target_branch, branch_level: $branch_level, from: $at, status: prop_rel_status }]->(prop_node)
        }
        CALL {
            WITH attr_rel, prop_rel_status, prop_type, prop_node
            WITH attr_rel, prop_rel_status, prop_type, prop_node
            WHERE prop_type = "IS_PROTECTED"
            CREATE (attr_rel)-[:IS_PROTECTED { branch: $target_branch, branch_level: $branch_level, from: $at, status: prop_rel_status }]->(prop_node)
        }
    }
}
        """
        self.add_to_query(query=query)


class DiffMergeRollbackQuery(Query):
    name = "diff_merge_rollback"
    type = QueryType.WRITE
    insert_return = False

    def __init__(
        self,
        at: Timestamp,
        target_branch: Branch,
        **kwargs: Any,
    ) -> None:
        super().__init__(**kwargs)
        self.at = at
        self.target_branch = target_branch
        self.source_branch_name = self.branch.name

    async def query_init(self, db: InfrahubDatabase, **kwargs: Any) -> None:  # noqa: ARG002
        self.params = {
            "at": self.at.to_string(),
            "target_branch": self.target_branch.name,
            "source_branch": self.source_branch_name,
        }
        query = """
        // ---------------------------
        // reset to times on target branch
        // ---------------------------
        CALL {
            OPTIONAL MATCH ()-[r_to {to: $at, branch: $target_branch}]-()
            SET r_to.to = NULL
        }
        // ---------------------------
        // reset from times on target branch
        // ---------------------------
        CALL {
            OPTIONAL MATCH ()-[r_from {from: $at, branch: $target_branch}]-()
            DELETE r_from
        }
        """
        self.add_to_query(query=query)



from pydantic import BaseModel, ConfigDict, Field

from infrahub.context import InfrahubContext
from infrahub.message_bus.messages.proposed_change.base_with_diff import BaseProposedChangeWithDiffMessage
from infrahub.message_bus.types import ProposedChangeArtifactDefinition, ProposedChangeBranchDiff


class RequestProposedChangeDataIntegrity(BaseProposedChangeWithDiffMessage):
    """Sent trigger data integrity checks for a proposed change"""


class RequestProposedChangeRunGenerators(BaseProposedChangeWithDiffMessage):
    """Sent trigger the generators that are impacted by the proposed change to run."""

    refresh_artifacts: bool = Field(..., description="Whether to regenerate artifacts after the generators are run")
    do_repository_checks: bool = Field(
        ..., description="Whether to run repository and user checks after the generators are run"
    )


class RequestProposedChangeRepositoryChecks(BaseProposedChangeWithDiffMessage):
    """Sent when a proposed change is created to trigger additional checks"""


class RequestProposedChangeSchemaIntegrity(BaseProposedChangeWithDiffMessage):
    """Sent trigger schema integrity checks for a proposed change"""


class RequestProposedChangeUserTests(BaseProposedChangeWithDiffMessage):
    """Sent trigger to run tests (smoke, units, integrations) for a proposed change."""


class RequestArtifactDefinitionCheck(BaseModel):
    """Sent to validate the generation of artifacts in relation to a proposed change."""

    model_config = ConfigDict(arbitrary_types_allowed=True)

    artifact_definition: ProposedChangeArtifactDefinition = Field(..., description="The Artifact Definition")
    branch_diff: ProposedChangeBranchDiff = Field(..., description="The calculated diff between the two branches")
    proposed_change: str = Field(..., description="The unique ID of the Proposed Change")
    source_branch: str = Field(..., description="The source branch")
    source_branch_sync_with_git: bool = Field(..., description="Indicates if the source branch should sync with git")
    destination_branch: str = Field(..., description="The target branch")

    context: InfrahubContext = Field(..., description="The context of the task")



from typing import Any

from infrahub import config
from infrahub.core.query import Query, QueryType
from infrahub.core.timestamp import Timestamp
from infrahub.database import InfrahubDatabase

from ..model.path import TrackingId
from .filters import EnrichedDiffQueryFilters

QUERY_MATCH_NODES = """
    // get the roots of all diffs in the query
    MATCH (diff_root:DiffRoot)
    WHERE (diff_root.is_merged IS NULL OR diff_root.is_merged <> TRUE)
    AND diff_root.base_branch = $base_branch
    AND diff_root.diff_branch IN $diff_branches
    AND ($from_time IS NULL OR diff_root.from_time >= $from_time)
    AND ($to_time IS NULL OR diff_root.to_time <= $to_time)
    AND ($tracking_id IS NULL OR diff_root.tracking_id = $tracking_id)
    AND ($diff_ids IS NULL OR diff_root.uuid IN $diff_ids)
    // get all the nodes attached to the diffs
    OPTIONAL MATCH (diff_root)-[:DIFF_HAS_NODE]->(diff_node:DiffNode)
    """


class EnrichedDiffGetQuery(Query):
    """Get all EnrichedDiffRoots for the given branches that are within the given timeframe in chronological order"""

    name = "enriched_diff_get"
    type = QueryType.READ
    insert_limit = False

    def __init__(
        self,
        base_branch_name: str,
        diff_branch_names: list[str],
        max_depth: int,
        filters: EnrichedDiffQueryFilters | None = None,
        from_time: Timestamp | None = None,
        to_time: Timestamp | None = None,
        tracking_id: TrackingId | None = None,
        diff_ids: list[str] | None = None,
        **kwargs: Any,
    ) -> None:
        super().__init__(**kwargs)
        self.base_branch_name = base_branch_name
        self.diff_branch_names = diff_branch_names
        self.from_time: Timestamp = from_time or Timestamp("2000-01-01T00:00:01Z")
        self.to_time: Timestamp = to_time or Timestamp()
        self.max_depth = max_depth
        self.tracking_id = tracking_id
        self.diff_ids = diff_ids
        self.filters = filters or EnrichedDiffQueryFilters()

    async def query_init(self, db: InfrahubDatabase, **kwargs: Any) -> None:  # noqa: ARG002
        self.params = {
            "base_branch": self.base_branch_name,
            "diff_branches": self.diff_branch_names,
            "from_time": self.from_time.to_string(),
            "to_time": self.to_time.to_string(),
            "tracking_id": self.tracking_id.serialize() if self.tracking_id else None,
            "diff_ids": self.diff_ids,
            "limit": self.limit or config.SETTINGS.database.query_size_limit,
            "offset": self.offset,
        }
        # ruff: noqa: E501
        self.add_to_query(query=QUERY_MATCH_NODES)

        if not self.filters.is_empty:
            filters, filter_params = self.filters.generate()
            self.params.update(filter_params)

            query_filters = """
            WHERE (
                %(filters)s
            )
            """ % {"filters": filters}
            self.add_to_query(query=query_filters)

        query_2 = """
        WITH diff_root, diff_node
        ORDER BY diff_root.base_branch, diff_root.diff_branch, diff_root.from_time, diff_root.to_time, diff_node.uuid
        // -------------------------------------
        // Limit number of results
        // -------------------------------------
        SKIP COALESCE($offset, 0)
        LIMIT $limit
        // -------------------------------------
        // Check if more data after this limited group
        // -------------------------------------
        WITH collect([diff_root, diff_node]) AS limited_results
        WITH limited_results, size(limited_results) = $limit AS has_more_data
        UNWIND limited_results AS one_result
        WITH one_result[0] AS diff_root, one_result[1] AS diff_node, has_more_data
        // if depth limit, make sure not to exceed it when traversing linked nodes
        // -------------------------------------
        // Retrieve Parents
        // -------------------------------------
        CALL {
            WITH diff_node
            OPTIONAL MATCH parents_path = (diff_node)-[:DIFF_HAS_RELATIONSHIP|DIFF_HAS_NODE*1..%(max_depth)s]->(:DiffNode)
            RETURN parents_path
            ORDER BY size(nodes(parents_path)) DESC
            LIMIT 1
        }
        WITH diff_root, diff_node, has_more_data, parents_path
        // -------------------------------------
        // Retrieve conflicts
        // -------------------------------------
        OPTIONAL MATCH (diff_node)-[:DIFF_HAS_CONFLICT]->(diff_node_conflict:DiffConflict)
        WITH diff_root, diff_node, has_more_data, parents_path, diff_node_conflict
        // -------------------------------------
        // Retrieve Attributes
        // -------------------------------------
        CALL {
            WITH diff_node
            OPTIONAL MATCH (diff_node)-[:DIFF_HAS_ATTRIBUTE]->(diff_attribute:DiffAttribute)
            WITH diff_attribute
            OPTIONAL MATCH (diff_attribute)-[:DIFF_HAS_PROPERTY]->(diff_attr_property:DiffProperty)
            WITH diff_attribute, diff_attr_property
            OPTIONAL MATCH (diff_attr_property)-[:DIFF_HAS_CONFLICT]->(diff_attr_property_conflict:DiffConflict)
            RETURN diff_attribute, diff_attr_property, diff_attr_property_conflict
            ORDER BY diff_attribute.name, diff_attr_property.property_type
        }
        WITH diff_root, diff_node, has_more_data, parents_path, diff_node_conflict, collect([diff_attribute, diff_attr_property, diff_attr_property_conflict]) as diff_attributes
        // -------------------------------------
        // Retrieve Relationships
        // -------------------------------------
        CALL {
            WITH diff_node
            OPTIONAL MATCH (diff_node)-[:DIFF_HAS_RELATIONSHIP]->(diff_relationship:DiffRelationship)
            WITH diff_relationship
            OPTIONAL MATCH (diff_relationship)-[:DIFF_HAS_ELEMENT]->(diff_rel_element:DiffRelationshipElement)
            WITH diff_relationship, diff_rel_element
            OPTIONAL MATCH (diff_rel_element)-[:DIFF_HAS_CONFLICT]->(diff_rel_conflict:DiffConflict)
            WITH diff_relationship, diff_rel_element, diff_rel_conflict
            OPTIONAL MATCH (diff_rel_element)-[:DIFF_HAS_PROPERTY]->(diff_rel_property:DiffProperty)
            WITH diff_relationship, diff_rel_element, diff_rel_conflict, diff_rel_property
            OPTIONAL MATCH (diff_rel_property)-[:DIFF_HAS_CONFLICT]->(diff_rel_property_conflict:DiffConflict)

            RETURN diff_relationship, diff_rel_element, diff_rel_conflict, diff_rel_property, diff_rel_property_conflict
            ORDER BY diff_relationship.name, diff_rel_element.peer_id, diff_rel_property.property_type
        }
        WITH
            diff_root,
            diff_node,
            has_more_data,
            parents_path,
            diff_node_conflict,
            diff_attributes,
            collect([diff_relationship, diff_rel_element, diff_rel_conflict, diff_rel_property, diff_rel_property_conflict]) AS diff_relationships
        """ % {"max_depth": self.max_depth * 2}

        self.add_to_query(query=query_2)

        self.return_labels = [
            "diff_root",
            "diff_node",
            "has_more_data",
            "parents_path",
            "diff_node_conflict",
            "diff_attributes",
            "diff_relationships",
        ]
        self.order_by = ["diff_root.diff_branch_name ASC", "diff_root.from_time ASC", "diff_node.label ASC"]



from __future__ import annotations

from copy import deepcopy
from typing import Any, ClassVar, Self, final
from uuid import UUID, uuid4

from pydantic import BaseModel, Field, PrivateAttr, model_validator

from infrahub import __version__
from infrahub.auth import AccountSession, AuthType
from infrahub.context import InfrahubContext  # noqa: TC001
from infrahub.core.branch import Branch  # noqa: TC001
from infrahub.message_bus import InfrahubMessage, Meta
from infrahub.worker import WORKER_IDENTITY

from .constants import EVENT_NAMESPACE


class EventNode(BaseModel):
    id: str
    kind: str


class ParentEvent(BaseModel):
    id: str
    name: str


class EventMeta(BaseModel):
    branch: Branch | None = Field(default=None, description="The branch on which originate this event")
    request_id: str = ""
    account_id: str | None = Field(default=None, description="The ID of the account triggering this event")
    initiator_id: str = Field(
        default=WORKER_IDENTITY, description="The worker identity of the initial sender of this message"
    )
    context: InfrahubContext = Field(..., description="The context used when originating this event")
    level: int = Field(default=0)
    has_children: bool = Field(
        default=False, description="Indicates if this event might potentially have child events under it."
    )

    id: UUID = Field(
        default_factory=uuid4,
        description="UUID of the event",
    )

    parent: UUID | None = Field(default=None, description="The UUID of the parent event if applicable")
    ancestors: list[ParentEvent] = Field(default_factory=list, description="Any event used to trigger this event")
    _created_with_context: bool = PrivateAttr(default=False)

    def get_branch_id(self) -> str:
        if self.context.branch.id:
            return self.context.branch.id

        if self.branch:
            return str(self.branch.get_uuid())

        return ""

    def get_id(self) -> str:
        return str(self.id)

    def get_related(self) -> list[dict[str, str]]:
        related: list[dict[str, str]] = [
            {"prefect.resource.id": __version__, "prefect.resource.role": "infrahub.version"},
            {
                "prefect.resource.id": self.get_id(),
                "prefect.resource.role": "infrahub.event",
                "infrahub.event.has_children": str(self.has_children).lower(),
                "infrahub.event.level": str(self.level),
            },
            {
                "prefect.resource.id": f"infrahub.account.{self.context.account.account_id}",
                "prefect.resource.role": "infrahub.account",
                "infrahub.resource.id": self.context.account.account_id,
            },
            {
                "prefect.resource.id": f"infrahub.branch.{self.get_branch_id()}",
                "prefect.resource.role": "infrahub.branch",
                "infrahub.resource.id": self.get_branch_id(),
                "infrahub.resource.label": self.context.branch.name,
            },
        ]

        if self.parent:
            related.append(
                {
                    "prefect.resource.id": self.get_id(),
                    "prefect.resource.role": "infrahub.child_event",
                    "infrahub.event_parent.id": str(self.parent),
                }
            )

        for ancestor in self.ancestors:
            related.append(
                {
                    "prefect.resource.id": ancestor.id,
                    "prefect.resource.role": "infrahub.ancestor_event",
                    "infrahub.ancestor_event.name": ancestor.name,
                }
            )

        return related

    @classmethod
    def with_dummy_context(cls, branch: Branch) -> EventMeta:
        return cls(
            branch=branch,
            context=InfrahubContext.init(
                branch=branch, account=AccountSession(auth_type=AuthType.NONE, authenticated=False, account_id="")
            ),
        )

    @classmethod
    def from_parent(cls, parent: InfrahubEvent, branch: Branch | None = None) -> EventMeta:
        """Create the metadata from an existing event

        Note that this action will modify the existing event to indicate that children might be attached to the event
        """
        parent.meta.has_children = True
        context = deepcopy(parent.meta.context)
        if branch:
            context.branch.name = branch.name
            context.branch.id = str(branch.get_uuid())

        return cls(
            parent=parent.meta.id,
            branch=parent.meta.branch,
            request_id=parent.meta.request_id,
            initiator_id=parent.meta.initiator_id,
            account_id=parent.meta.account_id,
            level=parent.meta.level + 1,
            context=context,
            ancestors=[ParentEvent(id=parent.get_id(), name=parent.event_name)] + parent.meta.ancestors,
        )

    @classmethod
    def from_context(cls, context: InfrahubContext, branch: Branch | None = None) -> EventMeta:
        # Create a copy of the context so local changes aren't brought back to a parent object
        meta = cls(context=deepcopy(context))
        meta._created_with_context = True
        if branch:
            meta.context.branch.name = branch.name
            meta.context.branch.id = str(branch.get_uuid())
        return meta


class InfrahubEvent(BaseModel):
    meta: EventMeta = Field(..., description="Metadata for the event")

    event_name: ClassVar[str] = Field(..., description="The name of the event")
    infrahub_node_kind_event: ClassVar[bool] = False

    def get_id(self) -> str:
        return self.meta.get_id()

    def get_event_namespace(self) -> str:
        return EVENT_NAMESPACE

    def get_resource(self) -> dict[str, str]:
        raise NotImplementedError

    def get_messages(self) -> list[InfrahubMessage]:
        return []

    def get_related(self) -> list[dict[str, str]]:
        if not self.meta:
            return []
        return self.meta.get_related()

    def get_payload(self) -> dict[str, Any]:
        """The purpose if this method is to allow subclasses to define their own payload.

        It should not be used to get the complete payload instead .get_event_payload() should
        be used for that as it will always contain the 'context' key regardless of changes
        in child classes
        """
        return self.model_dump(exclude={"meta"})

    @final
    def get_event_payload(self) -> dict[str, Any]:
        """This method should be used when emitting the event to the event broker"""
        event_payload = {"data": self.get_payload(), "context": self.meta.context.to_event()}
        return event_payload

    def get_message_meta(self) -> Meta:
        meta = Meta()

        meta.initiator_id = self.meta.initiator_id
        if self.meta.request_id:
            meta.initiator_id = self.meta.request_id

        return meta

    @model_validator(mode="after")
    def update_context(self) -> Self:
        """Update the context object using this event provided that the meta data was created with a context."""
        if self.meta._created_with_context:
            self.meta.context.set_event(self.event_name, id=self.get_id())
        return self



from infrahub.computed_attribute.triggers import (
    TRIGGER_COMPUTED_ATTRIBUTE_ALL_SCHEMA,
    TRIGGER_COMPUTED_ATTRIBUTE_PYTHON_SETUP_COMMIT,
)
from infrahub.trigger.models import TriggerDefinition
from infrahub.webhook.triggers import TRIGGER_WEBHOOK_DELETE, TRIGGER_WEBHOOK_SETUP_UPDATE

builtin_triggers: list[TriggerDefinition] = [
    TRIGGER_COMPUTED_ATTRIBUTE_ALL_SCHEMA,
    TRIGGER_COMPUTED_ATTRIBUTE_PYTHON_SETUP_COMMIT,
    TRIGGER_WEBHOOK_DELETE,
    TRIGGER_WEBHOOK_SETUP_UPDATE,
]



from __future__ import annotations

from collections import deque
from copy import deepcopy
from dataclasses import dataclass, field
from enum import Enum
from functools import cached_property
from typing import TYPE_CHECKING, Any

from graphql import (
    FieldNode,
    FragmentDefinitionNode,
    FragmentSpreadNode,
    GraphQLSchema,
    InlineFragmentNode,
    NamedTypeNode,
    NonNullTypeNode,
    OperationDefinitionNode,
    OperationType,
    SelectionSetNode,
)
from infrahub_sdk.analyzer import GraphQLQueryAnalyzer
from infrahub_sdk.utils import extract_fields

from infrahub.core.constants import RelationshipCardinality
from infrahub.core.schema import GenericSchema
from infrahub.exceptions import SchemaNotFoundError
from infrahub.graphql.utils import extract_schema_models

if TYPE_CHECKING:
    from infrahub.core.branch import Branch
    from infrahub.core.schema import MainSchemaTypes
    from infrahub.core.schema.schema_branch import SchemaBranch


class MutateAction(str, Enum):
    CREATE = "create"
    DELETE = "delete"
    UPDATE = "update"


class ContextType(str, Enum):
    EDGE = "edge"
    NODE = "node"
    DIRECT = "direct"
    OBJECT = "object"

    @classmethod
    def from_operation(cls, operation: OperationType) -> ContextType:
        match operation:
            case OperationType.QUERY:
                return cls.EDGE
            case OperationType.MUTATION:
                return cls.OBJECT
            case OperationType.SUBSCRIPTION:
                return cls.EDGE

    @classmethod
    def from_relationship_cardinality(cls, cardinality: RelationshipCardinality) -> ContextType:
        match cardinality:
            case RelationshipCardinality.MANY:
                return cls.EDGE
            case RelationshipCardinality.ONE:
                return cls.NODE


class GraphQLOperation(str, Enum):
    QUERY = "query"
    MUTATION = "mutation"
    SUBSCRIPTION = "subscription"
    UNDEFINED = "undefined"

    @classmethod
    def from_operation(cls, operation: OperationType) -> GraphQLOperation:
        match operation:
            case OperationType.QUERY:
                return cls.QUERY
            case OperationType.MUTATION:
                return cls.MUTATION
            case OperationType.SUBSCRIPTION:
                return cls.SUBSCRIPTION


@dataclass
class GraphQLSelectionSet:
    field_nodes: list[FieldNode]
    fragment_spread_nodes: list[FragmentSpreadNode]
    inline_fragment_nodes: list[InlineFragmentNode]


@dataclass
class GraphQLArgument:
    name: str
    value: str
    kind: str


@dataclass
class ObjectAccess:
    attributes: set[str] = field(default_factory=set)
    relationships: set[str] = field(default_factory=set)


@dataclass
class GraphQLVariable:
    name: str
    type: str
    required: bool


@dataclass
class GraphQLQueryModel:
    model: MainSchemaTypes
    root: bool
    arguments: list[GraphQLArgument]
    attributes: set[str]
    relationships: set[str]
    mutate_actions: list[MutateAction] = field(default_factory=list)


@dataclass
class GraphQLQueryNode:
    path: str
    operation: GraphQLOperation = field(default=GraphQLOperation.UNDEFINED)
    arguments: list[GraphQLArgument] = field(default_factory=list)
    variables: list[GraphQLVariable] = field(default_factory=list)
    context_type: ContextType = field(default=ContextType.EDGE)
    parent: GraphQLQueryNode | None = field(default=None)
    children: list[GraphQLQueryNode] = field(default_factory=list)
    infrahub_model: MainSchemaTypes | None = field(default=None)
    infrahub_node_models: list[MainSchemaTypes] = field(default_factory=list)
    infrahub_attributes: set[str] = field(default_factory=set)
    infrahub_relationships: set[str] = field(default_factory=set)
    field_node: FieldNode | None = field(default=None)
    mutate_actions: list[MutateAction] = field(default_factory=list)

    def context_model(self) -> MainSchemaTypes | None:
        """Return the closest Infrahub object by going up in the tree"""
        if self.infrahub_model:
            return self.infrahub_model
        if self.parent:
            return self.parent.context_model()

        return None

    def context_path(self) -> str:
        """Return the relative path for the current context with the closest Infrahub object as the root"""
        if self.infrahub_model:
            return f"/{self.path}"
        if self.parent:
            return f"{self.parent.context_path()}/{self.path}"
        return self.path

    def properties_path(self) -> str:
        """Indicate the expected path to where Infrahub attributes and relationships would be defined."""
        if self.infrahub_model:
            match self.context_type:
                case ContextType.DIRECT:
                    return f"/{self.path}"
                case ContextType.EDGE:
                    return f"/{self.path}/edges/node"
                case ContextType.NODE:
                    return f"/{self.path}/node"
                case ContextType.OBJECT:
                    return f"/{self.path}/object"
        if self.parent:
            return self.parent.properties_path()

        return self.path

    def full_path(self) -> str:
        """Return the full path within the tree for the current context."""
        if self.parent:
            return f"{self.parent.full_path()}/{self.path}"
        return self.path

    @property
    def at_root(self) -> bool:
        if self.parent:
            return False
        return True

    @property
    def in_property_level(self) -> bool:
        """Indicate if properties, i.e., attributes and relationships could exist at this level."""
        return self.context_path() == self.properties_path()

    def append_attribute(self, attribute: str) -> None:
        """Add attributes to the closes parent Infrahub object."""
        if self.infrahub_model:
            self.infrahub_attributes.add(attribute)
        elif self.parent:
            self.parent.append_attribute(attribute=attribute)

    def append_relationship(self, relationship: str) -> None:
        """Add relationships to the closes parent Infrahub object."""
        if self.infrahub_model:
            self.infrahub_relationships.add(relationship)
        elif self.parent:
            self.parent.append_relationship(relationship=relationship)

    def get_models(self) -> list[GraphQLQueryModel]:
        """Return all models defined on this node along with child nodes"""
        models: list[GraphQLQueryModel] = []
        if self.infrahub_model:
            models.append(
                GraphQLQueryModel(
                    model=self.infrahub_model,
                    root=self.at_root,
                    arguments=self.arguments,
                    attributes=self.infrahub_attributes,
                    relationships=self.infrahub_relationships,
                    mutate_actions=self.mutate_actions,
                )
            )
            for used_by in self.infrahub_node_models:
                models.append(
                    GraphQLQueryModel(
                        model=used_by,
                        root=self.at_root,
                        arguments=self.arguments,
                        attributes=self.infrahub_attributes,
                        relationships=self.infrahub_relationships,
                        mutate_actions=self.mutate_actions,
                    )
                )

        for child in self.children:
            models.extend(child.get_models())
        return models


@dataclass
class GraphQLQueryReport:
    queries: list[GraphQLQueryNode]

    @property
    def impacted_models(self) -> list[str]:
        """Return a list of all Infrahub objects that are impacted by queries within the request"""
        models: set[str] = set()
        for query in self.queries:
            query_models = query.get_models()
            models.update([query_model.model.kind for query_model in query_models])

        return sorted(models)

    @cached_property
    def requested_read(self) -> dict[str, ObjectAccess]:
        """Return Infrahub objects and the fields (attributes and relationships) that this query would attempt to read"""
        access: dict[str, ObjectAccess] = {}
        for query in self.queries:
            query_models = query.get_models()
            for query_model in query_models:
                if query_model.model.kind not in access:
                    access[query_model.model.kind] = ObjectAccess()
                access[query_model.model.kind].attributes.update(query_model.attributes)
                access[query_model.model.kind].relationships.update(query_model.relationships)

        return access

    def fields_by_kind(self, kind: str) -> list[str]:
        fields: list[str] = []
        if access := self.requested_read.get(kind):
            fields.extend(list(access.attributes))
            fields.extend(list(access.relationships))

        return fields

    @cached_property
    def top_level_kinds(self) -> list[str]:
        return [query.infrahub_model.kind for query in self.queries if query.infrahub_model]

    @cached_property
    def kind_action_map(self) -> dict[str, set[MutateAction]]:
        access: dict[str, set[MutateAction]] = {}
        root_models: set[str] = set()
        includes_mutations: bool = False
        for query in self.queries:
            query_models = query.get_models()
            for query_model in query_models:
                if query_model.mutate_actions:
                    includes_mutations = True
                if includes_mutations:
                    if query_model.model.kind not in access:
                        access[query_model.model.kind] = set()
                    if query_model.root:
                        root_models.add(query_model.model.kind)
                    access[query_model.model.kind].update(query_model.mutate_actions)

        # Until we properly analyze the data payload it is assumed that the required permissions on non-root objects is update
        # the idea around this is that at this point even if we only return data from a relationship without actually updating
        # that relationship we'd still expect to have UPDATE permissions on that related object. However, this is still a step
        # in the right direction as we'd previously require the same permissions as that of the base object so this is still
        # more correct.
        for node_kind, node_actions in access.items():
            if node_kind not in root_models:
                node_actions.add(MutateAction.UPDATE)

        return access


class InfrahubGraphQLQueryAnalyzer(GraphQLQueryAnalyzer):
    def __init__(
        self,
        query: str,
        branch: Branch,
        schema_branch: SchemaBranch,
        schema: GraphQLSchema | None = None,
        query_variables: dict[str, Any] | None = None,
        operation_name: str | None = None,
    ) -> None:
        self.branch = branch
        self.schema_branch = schema_branch
        self.operation_name = operation_name
        self.query_variables: dict[str, Any] = query_variables or {}
        self._named_fragments: dict[str, GraphQLQueryNode] = {}
        self._fragment_dependencies: dict[str, set[str]] = {}
        super().__init__(query=query, schema=schema)

    @property
    def operation_names(self) -> list[str]:
        return [operation.name for operation in self.operations if operation.name is not None]

    @cached_property
    def _fragment_definitions(self) -> list[FragmentDefinitionNode]:
        return [
            definition for definition in self.document.definitions if isinstance(definition, FragmentDefinitionNode)
        ]

    @cached_property
    def _operation_definitions(self) -> list[OperationDefinitionNode]:
        return [
            definition for definition in self.document.definitions if isinstance(definition, OperationDefinitionNode)
        ]

    def get_named_fragment_with_parent(self, name: str, parent: GraphQLQueryNode) -> GraphQLQueryNode:
        """Return a copy of the named fragment and attach it to a parent.

        We return a copy of the object as a named fragment could be used by multiple queries and as we're
        generally working with references to objects we wouldn't want to override the parent of a previously
        assigned object
        """
        named_fragment = deepcopy(self._named_fragments[name])
        named_fragment.parent = parent
        return named_fragment

    async def get_models_in_use(self, types: dict[str, Any]) -> set[str]:
        """List of Infrahub models that are referenced in the query."""
        graphql_types = set()
        models = set()

        if not self.schema:
            raise ValueError("Schema must be provided to extract the models in use.")

        for definition in self.document.definitions:
            fields = await extract_fields(definition.selection_set)

            operation = getattr(definition, "operation", None)
            if operation == OperationType.QUERY:
                schema = self.schema.query_type
            elif operation == OperationType.MUTATION:
                schema = self.schema.mutation_type
            else:
                # Subscription not supported right now
                continue

            graphql_types.update(await extract_schema_models(fields=fields, schema=schema, root_schema=self.schema))

        for graphql_type_name in graphql_types:
            try:
                graphql_type = types.get(graphql_type_name)
                if not hasattr(graphql_type, "_meta") or not hasattr(graphql_type._meta, "schema"):  # type: ignore[union-attr]
                    continue
                models.add(graphql_type._meta.schema.kind)  # type: ignore[union-attr]
            except ValueError:
                continue

        return models

    @cached_property
    def query_report(self) -> GraphQLQueryReport:
        self._populate_named_fragments()
        operations = self._get_operations()

        return GraphQLQueryReport(queries=operations)

    def _get_operations(self) -> list[GraphQLQueryNode]:
        operations: list[GraphQLQueryNode] = []
        for operation_definition in self._operation_definitions:
            selections = self._get_selections(selection_set=operation_definition.selection_set)

            for field_node in selections.field_nodes:
                schema_model: MainSchemaTypes
                infrahub_node_models: list[MainSchemaTypes] = []
                model_name = self._get_model_name(node=field_node, operation_definition=operation_definition)

                if model_name in self.schema_branch.node_names:
                    schema_model = self.schema_branch.get_node(name=model_name, duplicate=False)
                elif model_name in self.schema_branch.generic_names:
                    schema_model = self.schema_branch.get_generic(name=model_name, duplicate=False)
                    infrahub_node_models = [
                        self.schema_branch.get(name=used_by, duplicate=False) for used_by in schema_model.used_by
                    ]
                elif model_name in self.schema_branch.profile_names:
                    schema_model = self.schema_branch.get_profile(name=model_name, duplicate=False)
                else:
                    continue

                operational_node = GraphQLQueryNode(
                    operation=GraphQLOperation.from_operation(operation=operation_definition.operation),
                    path=schema_model.kind,
                    infrahub_model=schema_model,
                    infrahub_node_models=infrahub_node_models,
                    mutate_actions=self._get_model_mutations(
                        node=field_node, operation_definition=operation_definition
                    ),
                    context_type=ContextType.from_operation(operation=operation_definition.operation),
                    arguments=self._parse_arguments(field_node=field_node),
                    variables=self._get_variables(operation=operation_definition),
                )

                if field_node.selection_set:
                    selections = self._get_selections(selection_set=field_node.selection_set)
                    for selection_field_node in selections.field_nodes:
                        operational_node.children.append(
                            self._populate_field_node(node=selection_field_node, query_node=operational_node)
                        )
                operations.append(operational_node)
        return operations

    @staticmethod
    def _get_model_name(node: FieldNode, operation_definition: OperationDefinitionNode) -> str:
        if operation_definition.operation == OperationType.MUTATION and node.name.value.endswith(
            ("Create", "Delete", "Update", "Upsert")
        ):
            return node.name.value[:-6]
        return node.name.value

    @staticmethod
    def _get_model_mutations(node: FieldNode, operation_definition: OperationDefinitionNode) -> list[MutateAction]:
        if operation_definition.operation == OperationType.MUTATION:
            if node.name.value.endswith("Create"):
                return [MutateAction.CREATE]
            if node.name.value.endswith("Delete"):
                return [MutateAction.DELETE]
            if node.name.value.endswith("Update"):
                return [MutateAction.UPDATE]
            if node.name.value.endswith("Upsert"):
                return [MutateAction.CREATE, MutateAction.UPDATE]
        return []

    @property
    def _sorted_fragment_definitions(self) -> list[FragmentDefinitionNode]:
        """Sort fragments so that we start processing fragments that don't depend on other fragments"""
        dependencies = deepcopy(self._fragment_dependencies)

        independent_fragments = deque([frag for frag, deps in dependencies.items() if not deps])

        sorted_fragments = []

        while independent_fragments:
            fragment_name = independent_fragments.popleft()
            sorted_fragments.append(fragment_name)

            for dependent, deps in dependencies.items():
                if fragment_name in deps:
                    deps.remove(fragment_name)
                    if not deps:
                        independent_fragments.append(dependent)

        if len(sorted_fragments) != len(self._fragment_dependencies):
            raise ValueError("Circular fragment dependency detected.")

        fragment_name_to_definition = {frag.name.value: frag for frag in self._fragment_definitions}
        return [fragment_name_to_definition[name] for name in sorted_fragments]

    def _populate_fragment_dependency(self, name: str, selection_set: SelectionSetNode | None) -> None:
        if selection_set:
            for selection in selection_set.selections:
                if isinstance(selection, FragmentSpreadNode):
                    self._fragment_dependencies[name].add(selection.name.value)
                elif isinstance(selection, FieldNode):
                    self._populate_fragment_dependency(name=name, selection_set=selection.selection_set)
                elif isinstance(selection, InlineFragmentNode):
                    self._populate_fragment_dependency(name=name, selection_set=selection.selection_set)

    def _populate_fragment_dependencies(self) -> None:
        for fragment in self._fragment_definitions:
            fragment_name = fragment.name.value
            self._fragment_dependencies[fragment_name] = set()
            self._populate_fragment_dependency(name=fragment_name, selection_set=fragment.selection_set)

    def _populate_named_fragments(self) -> None:
        self._populate_fragment_dependencies()
        self._named_fragments = {}

        for fragment_definition in self._sorted_fragment_definitions:
            fragment_name = fragment_definition.name.value
            condition_name = fragment_definition.type_condition.name.value
            selections = self._get_selections(selection_set=fragment_definition.selection_set)

            try:
                infrahub_model = self.schema_branch.get(name=condition_name, duplicate=False)
            except SchemaNotFoundError:
                infrahub_model = None

            named_fragment = GraphQLQueryNode(
                path=fragment_definition.type_condition.name.value,
                context_type=ContextType.DIRECT,
                infrahub_model=infrahub_model,
            )
            for field_node in selections.field_nodes:
                named_fragment.children.append(self._populate_field_node(node=field_node, query_node=named_fragment))
            for inline_fragment_node in selections.inline_fragment_nodes:
                named_fragment.children.append(
                    self._populate_inline_fragment_node(node=inline_fragment_node, query_node=named_fragment)
                )

            self._named_fragments[fragment_name] = named_fragment

    def _populate_field_node(self, node: FieldNode, query_node: GraphQLQueryNode) -> GraphQLQueryNode:
        context_type = query_node.context_type
        infrahub_model = None
        infrahub_node_models: list[MainSchemaTypes] = []
        if query_node.in_property_level:
            if model := query_node.context_model():
                if node.name.value in model.attribute_names or node.name.value == "display_label":
                    query_node.append_attribute(attribute=node.name.value)
                elif node.name.value in model.relationship_names:
                    rel = model.get_relationship_or_none(name=node.name.value)
                    if rel:
                        infrahub_model = self.schema_branch.get(name=rel.peer, duplicate=False)
                        if isinstance(infrahub_model, GenericSchema):
                            infrahub_node_models = [
                                self.schema_branch.get(name=used_by, duplicate=False)
                                for used_by in infrahub_model.used_by
                            ]

                        context_type = ContextType.from_relationship_cardinality(cardinality=rel.cardinality)
                    query_node.append_relationship(relationship=node.name.value)

        current_node = GraphQLQueryNode(
            parent=query_node,
            path=node.name.value,
            context_type=context_type,
            infrahub_model=infrahub_model,
            infrahub_node_models=infrahub_node_models,
            arguments=self._parse_arguments(field_node=node),
        )

        if node.selection_set:
            selections = self._get_selections(selection_set=node.selection_set)
            for field_node in selections.field_nodes:
                current_node.children.append(self._populate_field_node(node=field_node, query_node=current_node))
            for inline_fragment_node in selections.inline_fragment_nodes:
                current_node.children.append(
                    self._populate_inline_fragment_node(node=inline_fragment_node, query_node=current_node)
                )
            for fragment_spread_node in selections.fragment_spread_nodes:
                current_node.children.append(
                    self._populate_fragment_spread_node(node=fragment_spread_node, query_node=current_node)
                )

        return current_node

    def _populate_inline_fragment_node(
        self, node: InlineFragmentNode, query_node: GraphQLQueryNode
    ) -> GraphQLQueryNode:
        context_type = query_node.context_type
        infrahub_model = self.schema_branch.get(name=node.type_condition.name.value)
        context_type = ContextType.DIRECT
        current_node = GraphQLQueryNode(
            parent=query_node,
            path=node.type_condition.name.value,
            context_type=context_type,
            infrahub_model=infrahub_model,
        )
        if node.selection_set:
            selections = self._get_selections(selection_set=node.selection_set)
            for field_node in selections.field_nodes:
                current_node.children.append(self._populate_field_node(node=field_node, query_node=current_node))
            for inline_fragment_node in selections.inline_fragment_nodes:
                current_node.children.append(
                    self._populate_inline_fragment_node(node=inline_fragment_node, query_node=current_node)
                )

        return current_node

    def _populate_fragment_spread_node(
        self, node: FragmentSpreadNode, query_node: GraphQLQueryNode
    ) -> GraphQLQueryNode:
        return self.get_named_fragment_with_parent(name=node.name.value, parent=query_node)

    @staticmethod
    def _get_selections(selection_set: SelectionSetNode) -> GraphQLSelectionSet:
        return GraphQLSelectionSet(
            field_nodes=[selection for selection in selection_set.selections if isinstance(selection, FieldNode)],
            fragment_spread_nodes=[
                selection for selection in selection_set.selections if isinstance(selection, FragmentSpreadNode)
            ],
            inline_fragment_nodes=[
                selection for selection in selection_set.selections if isinstance(selection, InlineFragmentNode)
            ],
        )

    @staticmethod
    def _get_variables(operation: OperationDefinitionNode) -> list[GraphQLVariable]:
        variables = []
        for variable in operation.variable_definitions:
            if isinstance(variable.type, NamedTypeNode):
                variables.append(
                    GraphQLVariable(name=variable.variable.name.value, type=variable.type.name.value, required=False)
                )
            elif isinstance(variable.type, NonNullTypeNode):
                if isinstance(variable.type.type, NamedTypeNode):
                    variables.append(
                        GraphQLVariable(
                            name=variable.variable.name.value, type=variable.type.type.name.value, required=True
                        )
                    )

        return variables

    @staticmethod
    def _parse_arguments(field_node: FieldNode) -> list[GraphQLArgument]:
        return [
            GraphQLArgument(
                name=argument.name.value,
                value=getattr(argument.value, "value", ""),
                kind=argument.value.kind,
            )
            for argument in field_node.arguments
        ]



from __future__ import annotations

from typing import Any
from uuid import uuid4

import pydantic
from prefect import flow, get_run_logger
from prefect.client.schemas.objects import State  # noqa: TC002
from prefect.states import Completed, Failed

from infrahub import lock
from infrahub.context import InfrahubContext  # noqa: TC001  needed for prefect flow
from infrahub.core import registry
from infrahub.core.branch import Branch
from infrahub.core.changelog.diff import DiffChangelogCollector, MigrationTracker
from infrahub.core.constants import MutationAction
from infrahub.core.diff.coordinator import DiffCoordinator
from infrahub.core.diff.ipam_diff_parser import IpamDiffParser
from infrahub.core.diff.merger.merger import DiffMerger
from infrahub.core.diff.model.path import BranchTrackingId, EnrichedDiffRoot, EnrichedDiffRootMetadata
from infrahub.core.diff.models import RequestDiffUpdate
from infrahub.core.diff.repository.repository import DiffRepository
from infrahub.core.merge import BranchMerger
from infrahub.core.migrations.schema.models import SchemaApplyMigrationData
from infrahub.core.migrations.schema.tasks import schema_apply_migrations
from infrahub.core.timestamp import Timestamp
from infrahub.core.validators.determiner import ConstraintValidatorDeterminer
from infrahub.core.validators.models.validate_migration import SchemaValidateMigrationData
from infrahub.core.validators.tasks import schema_validate_migrations
from infrahub.dependencies.registry import get_component_registry
from infrahub.events.branch_action import BranchCreatedEvent, BranchDeletedEvent, BranchMergedEvent, BranchRebasedEvent
from infrahub.events.models import EventMeta, InfrahubEvent
from infrahub.events.node_action import get_node_event
from infrahub.exceptions import BranchNotFoundError, MergeFailedError, ValidationError
from infrahub.graphql.mutations.models import BranchCreateModel  # noqa: TC001
from infrahub.services import InfrahubServices  # noqa: TC001  needed for prefect flow
from infrahub.workflows.catalogue import (
    BRANCH_CANCEL_PROPOSED_CHANGES,
    BRANCH_MERGE_POST_PROCESS,
    DIFF_REFRESH_ALL,
    DIFF_UPDATE,
    GIT_REPOSITORIES_CREATE_BRANCH,
    IPAM_RECONCILIATION,
    TRIGGER_ARTIFACT_DEFINITION_GENERATE,
    TRIGGER_GENERATOR_DEFINITION_RUN,
)
from infrahub.workflows.utils import add_tags


@flow(name="branch-rebase", flow_run_name="Rebase branch {branch}")
async def rebase_branch(branch: str, context: InfrahubContext, service: InfrahubServices) -> None:  # noqa: PLR0915
    async with service.database.start_session() as db:
        log = get_run_logger()
        await add_tags(branches=[branch])
        obj = await Branch.get_by_name(db=db, name=branch)
        base_branch = await Branch.get_by_name(db=db, name=registry.default_branch)
        component_registry = get_component_registry()
        diff_repository = await component_registry.get_component(DiffRepository, db=db, branch=obj)
        diff_coordinator = await component_registry.get_component(DiffCoordinator, db=db, branch=obj)
        diff_merger = await component_registry.get_component(DiffMerger, db=db, branch=obj)
        initial_from_time = Timestamp(obj.get_branched_from())
        merger = BranchMerger(
            db=db,
            diff_coordinator=diff_coordinator,
            diff_merger=diff_merger,
            diff_repository=diff_repository,
            source_branch=obj,
            service=service,
        )

        enriched_diff_metadata = await diff_coordinator.update_branch_diff(base_branch=base_branch, diff_branch=obj)
        async for _ in diff_repository.get_all_conflicts_for_diff(
            diff_branch_name=enriched_diff_metadata.diff_branch_name, diff_id=enriched_diff_metadata.uuid
        ):
            # if there are any conflicts, raise the error
            raise ValidationError(
                f"Branch {obj.name} contains conflicts with the default branch that must be addressed."
                " Please review the diff for details and manually update the conflicts before rebasing."
            )
        node_diff_field_summaries = await diff_repository.get_node_field_summaries(
            diff_branch_name=enriched_diff_metadata.diff_branch_name, diff_id=enriched_diff_metadata.uuid
        )

        candidate_schema = merger.get_candidate_schema()
        determiner = ConstraintValidatorDeterminer(schema_branch=candidate_schema)
        constraints = await determiner.get_constraints(node_diffs=node_diff_field_summaries)

        # If there are some changes related to the schema between this branch and main, we need to
        #  - Run all the validations to ensure everything is correct before rebasing the branch
        #  - Run all the migrations after the rebase
        if obj.has_schema_changes:
            constraints += await merger.calculate_validations(target_schema=candidate_schema)
        if constraints:
            responses = await schema_validate_migrations(
                message=SchemaValidateMigrationData(
                    branch=obj, schema_branch=candidate_schema, constraints=constraints
                ),
                service=service,
            )
            error_messages = [violation.message for response in responses for violation in response.violations]
            if error_messages:
                raise ValidationError(",\n".join(error_messages))

        schema_in_main_before = merger.destination_schema.duplicate()
        migrations = []
        async with lock.registry.global_graph_lock():
            async with db.start_transaction() as dbt:
                await obj.rebase(db=dbt)
                log.info("Branch successfully rebased")

            if obj.has_schema_changes:
                # NOTE there is a bit additional work in order to calculate a proper diff that will
                # allow us to pull only the part of the schema that has changed, for now the safest option is to pull
                # Everything
                # schema_diff = await merger.has_schema_changes()
                # TODO Would be good to convert this part to a Prefect Task in order to track it properly
                updated_schema = await registry.schema.load_schema_from_db(
                    db=db,
                    branch=obj,
                    # schema=merger.source_schema.duplicate(),
                    # schema_diff=schema_diff,
                )
                registry.schema.set_schema_branch(name=obj.name, schema=updated_schema)
                obj.update_schema_hash()
                await obj.save(db=db)

                # Execute the migrations
                migrations = await merger.calculate_migrations(target_schema=updated_schema)

                errors = await schema_apply_migrations(
                    message=SchemaApplyMigrationData(
                        branch=merger.source_branch,
                        new_schema=candidate_schema,
                        previous_schema=schema_in_main_before,
                        migrations=migrations,
                    ),
                    service=service,
                )
                for error in errors:
                    log.error(error)

        default_branch_diff = await _get_diff_root(
            diff_coordinator=diff_coordinator,
            enriched_diff_metadata=enriched_diff_metadata,
            diff_repository=diff_repository,
            base_branch=base_branch,
            target_from=initial_from_time,
        )

        # -------------------------------------------------------------
        # Trigger the reconciliation of IPAM data after the rebase
        # -------------------------------------------------------------
        diff_parser = await component_registry.get_component(IpamDiffParser, db=db, branch=obj)
        ipam_node_details = await diff_parser.get_changed_ipam_node_details(
            source_branch_name=obj.name,
            target_branch_name=registry.default_branch,
        )
        if ipam_node_details:
            await service.workflow.submit_workflow(
                workflow=IPAM_RECONCILIATION,
                context=context,
                parameters={"branch": obj.name, "ipam_node_details": ipam_node_details},
            )

    await service.workflow.submit_workflow(
        workflow=DIFF_REFRESH_ALL, context=context, parameters={"branch_name": obj.name}
    )

    # -------------------------------------------------------------
    # Generate an event to indicate that a branch has been rebased
    # -------------------------------------------------------------
    rebase_event = BranchRebasedEvent(
        branch_name=obj.name, branch_id=str(obj.uuid), meta=EventMeta(branch=obj, context=context)
    )
    events: list[InfrahubEvent] = [rebase_event]
    changelog_collector = DiffChangelogCollector(
        diff=default_branch_diff, branch=obj, db=db, migration_tracker=MigrationTracker(migrations=migrations)
    )
    for action, node_changelog in changelog_collector.collect_changelogs():
        node_event_class = get_node_event(MutationAction.from_diff_action(diff_action=action))
        mutate_event = node_event_class(
            kind=node_changelog.node_kind,
            node_id=node_changelog.node_id,
            changelog=node_changelog,
            fields=node_changelog.updated_fields,
            meta=EventMeta.from_parent(parent=rebase_event, branch=obj),
        )
        events.append(mutate_event)

    for event in events:
        await service.event.send(event)


@flow(name="branch-merge", flow_run_name="Merge branch {branch} into main")
async def merge_branch(
    branch: str, context: InfrahubContext, service: InfrahubServices, proposed_change_id: str | None = None
) -> None:
    async with service.database.start_session() as db:
        log = get_run_logger()

        await add_tags(branches=[branch, registry.default_branch])

        obj = await Branch.get_by_name(db=db, name=branch)
        default_branch = await registry.get_branch(db=db, branch=registry.default_branch)
        component_registry = get_component_registry()
        merge_event = BranchMergedEvent(
            branch_name=obj.name,
            branch_id=str(obj.get_uuid()),
            proposed_change_id=proposed_change_id,
            meta=EventMeta.from_context(context=context, branch=registry.get_global_branch()),
        )

        merger: BranchMerger | None = None
        async with lock.registry.global_graph_lock():
            # await update_diff(model=RequestDiffUpdate(branch_name=obj.name))

            diff_repository = await component_registry.get_component(DiffRepository, db=db, branch=obj)
            diff_coordinator = await component_registry.get_component(DiffCoordinator, db=db, branch=obj)
            diff_merger = await component_registry.get_component(DiffMerger, db=db, branch=obj)
            merger = BranchMerger(
                db=db,
                diff_coordinator=diff_coordinator,
                diff_merger=diff_merger,
                diff_repository=diff_repository,
                source_branch=obj,
                service=service,
            )
            try:
                branch_diff = await merger.merge()
            except Exception as exc:
                log.exception("Merge failed, beginning rollback")
                await merger.rollback()
                raise MergeFailedError(branch_name=branch) from exc
            await merger.update_schema()

        changelog_collector = DiffChangelogCollector(diff=branch_diff, branch=obj, db=db)
        node_events = changelog_collector.collect_changelogs()
        if merger and merger.migrations:
            errors = await schema_apply_migrations(
                message=SchemaApplyMigrationData(
                    branch=merger.destination_branch,
                    new_schema=merger.destination_schema,
                    previous_schema=merger.initial_source_schema,
                    migrations=merger.migrations,
                ),
                service=service,
            )
            for error in errors:
                log.error(error)

        # -------------------------------------------------------------
        # Trigger the reconciliation of IPAM data after the merge
        # -------------------------------------------------------------
        diff_parser = await component_registry.get_component(IpamDiffParser, db=db, branch=obj)
        ipam_node_details = await diff_parser.get_changed_ipam_node_details(
            source_branch_name=obj.name,
            target_branch_name=registry.default_branch,
        )
        if ipam_node_details:
            await service.workflow.submit_workflow(
                workflow=IPAM_RECONCILIATION,
                context=context,
                parameters={"branch": registry.default_branch, "ipam_node_details": ipam_node_details},
            )
        # -------------------------------------------------------------
        # remove tracking ID from the diff because there is no diff after the merge
        # -------------------------------------------------------------
        diff_repository = await component_registry.get_component(DiffRepository, db=db, branch=obj)
        await diff_repository.mark_tracking_ids_merged(tracking_ids=[BranchTrackingId(name=obj.name)])

        # -------------------------------------------------------------
        # Generate an event to indicate that a branch has been merged
        # NOTE: we still need to convert this event and potentially pull
        #   some tasks currently executed based on the event into this workflow
        # -------------------------------------------------------------
        await service.workflow.submit_workflow(
            workflow=BRANCH_MERGE_POST_PROCESS,
            context=context,
            parameters={"source_branch": obj.name, "target_branch": registry.default_branch},
        )

        events: list[InfrahubEvent] = [merge_event]

        for action, node_changelog in node_events:
            meta = EventMeta.from_parent(parent=merge_event, branch=default_branch)
            node_event_class = get_node_event(MutationAction.from_diff_action(diff_action=action))
            mutate_event = node_event_class(
                kind=node_changelog.node_kind,
                node_id=node_changelog.node_id,
                changelog=node_changelog,
                fields=node_changelog.updated_fields,
                meta=meta,
            )
            events.append(mutate_event)

        for event in events:
            await service.event.send(event=event)


@flow(name="branch-delete", flow_run_name="Delete branch {branch}")
async def delete_branch(branch: str, context: InfrahubContext, service: InfrahubServices) -> None:
    await add_tags(branches=[branch])

    async with service.database.start_session() as db:
        obj = await Branch.get_by_name(db=db, name=str(branch))
        await obj.delete(db=db)

        event = BranchDeletedEvent(
            branch_name=branch,
            branch_id=str(obj.uuid),
            sync_with_git=obj.sync_with_git,
            meta=EventMeta.from_context(context=context, branch=registry.get_global_branch()),
        )

        await service.workflow.submit_workflow(
            workflow=BRANCH_CANCEL_PROPOSED_CHANGES, context=context, parameters={"branch_name": branch}
        )

        await service.event.send(event=event)


@flow(
    name="branch-validate",
    flow_run_name="Validate branch {branch} for conflicts",
    description="Validate if the branch has some conflicts",
    persist_result=True,
)
async def validate_branch(branch: str, service: InfrahubServices) -> State:
    await add_tags(branches=[branch])

    async with service.database.start_session() as db:
        obj = await Branch.get_by_name(db=db, name=branch)

        component_registry = get_component_registry()
        diff_repo = await component_registry.get_component(DiffRepository, db=db, branch=obj)
        has_conflicts = await diff_repo.diff_has_conflicts(
            diff_branch_name=obj.name, tracking_id=BranchTrackingId(name=obj.name)
        )
        if has_conflicts:
            return Failed(message="branch has some conflicts")
        return Completed(message="branch is valid")


@flow(name="create-branch", flow_run_name="Create branch {model.name}")
async def create_branch(model: BranchCreateModel, context: InfrahubContext, service: InfrahubServices) -> None:
    await add_tags(branches=[model.name])

    async with service.database.start_session() as db:
        try:
            await Branch.get_by_name(db=db, name=model.name)
            raise ValueError(f"The branch {model.name}, already exist")
        except BranchNotFoundError:
            pass

        data_dict: dict[str, Any] = dict(model)
        data_dict.pop("is_isolated", None)

        try:
            obj = Branch(**data_dict)
        except pydantic.ValidationError as exc:
            error_msgs = [f"invalid field {error['loc'][0]}: {error['msg']}" for error in exc.errors()]
            raise ValueError("\n".join(error_msgs)) from exc

        async with lock.registry.local_schema_lock():
            # Copy the schema from the origin branch and set the hash and the schema_changed_at value
            origin_schema = registry.schema.get_schema_branch(name=obj.origin_branch)
            new_schema = origin_schema.duplicate(name=obj.name)
            registry.schema.set_schema_branch(name=obj.name, schema=new_schema)
            obj.update_schema_hash()
            await obj.save(db=db)

            # Add Branch to registry
            registry.branch[obj.name] = obj
            await service.component.refresh_schema_hash(branches=[obj.name])

        event = BranchCreatedEvent(
            branch_name=obj.name,
            branch_id=str(obj.uuid),
            sync_with_git=obj.sync_with_git,
            meta=EventMeta.from_context(context=context, branch=registry.get_global_branch()),
        )
        await service.event.send(event=event)

        if obj.sync_with_git:
            await service.workflow.submit_workflow(
                workflow=GIT_REPOSITORIES_CREATE_BRANCH,
                context=context,
                parameters={"branch": obj.name, "branch_id": str(obj.uuid)},
            )


async def _get_diff_root(
    diff_coordinator: DiffCoordinator,
    enriched_diff_metadata: EnrichedDiffRootMetadata,
    diff_repository: DiffRepository,
    base_branch: Branch,
    target_from: Timestamp,
) -> EnrichedDiffRoot:
    default_branch_diff = await diff_coordinator.create_or_update_arbitrary_timeframe_diff(
        base_branch=base_branch,
        diff_branch=base_branch,
        from_time=target_from,
        to_time=enriched_diff_metadata.to_time,
        name=str(uuid4()),
    )
    # make sure we have the actual diff with data and not just the metadata
    if not isinstance(default_branch_diff, EnrichedDiffRoot):
        default_branch_diff = await diff_repository.get_one(
            diff_branch_name=base_branch.name, diff_id=default_branch_diff.uuid
        )

    return default_branch_diff


@flow(
    name="branch-merge-post-process",
    flow_run_name="Run additional tasks after merging {source_branch} in {target_branch}",
)
async def post_process_branch_merge(
    source_branch: str, target_branch: str, context: InfrahubContext, service: InfrahubServices
) -> None:
    async with service.database.start_session() as db:
        await add_tags(branches=[source_branch])
        log = get_run_logger()
        log.info(f"Running additional tasks after merging {source_branch} within {target_branch}")

        component_registry = get_component_registry()
        default_branch = registry.get_branch_from_registry()
        diff_repository = await component_registry.get_component(DiffRepository, db=db, branch=default_branch)
        # send diff update requests for every branch-tracking diff
        branch_diff_roots = await diff_repository.get_roots_metadata(base_branch_names=[target_branch])

        await service.workflow.submit_workflow(
            workflow=TRIGGER_ARTIFACT_DEFINITION_GENERATE,
            context=context,
            parameters={"branch": target_branch},
        )

        await service.workflow.submit_workflow(
            workflow=TRIGGER_GENERATOR_DEFINITION_RUN,
            context=context,
            parameters={"branch": target_branch},
        )

        for diff_root in branch_diff_roots:
            if (
                diff_root.base_branch_name != diff_root.diff_branch_name
                and diff_root.tracking_id
                and isinstance(diff_root.tracking_id, BranchTrackingId)
            ):
                request_diff_update_model = RequestDiffUpdate(branch_name=diff_root.diff_branch_name)
                await service.workflow.submit_workflow(
                    workflow=DIFF_UPDATE, context=context, parameters={"model": request_diff_update_model}
                )



from typing import Any

from infrahub.core.query import Query, QueryType
from infrahub.database import InfrahubDatabase

from ..model.path import BranchTrackingId


class EnrichedDiffHasConflictQuery(Query):
    name = "enriched_diff_has_conflicts"
    type = QueryType.READ
    insert_return = False

    def __init__(
        self,
        diff_branch_name: str,
        tracking_id: BranchTrackingId | None = None,
        diff_id: str | None = None,
        **kwargs: Any,
    ) -> None:
        super().__init__(**kwargs)
        if (diff_id is None and tracking_id is None) or (diff_id and tracking_id):
            raise ValueError("EnrichedDiffHasConflictQuery requires one and only one of `tracking_id` or `diff_id`")
        self.diff_branch_name = diff_branch_name
        self.tracking_id = tracking_id
        self.diff_id = diff_id

    async def query_init(self, db: InfrahubDatabase, **kwargs: Any) -> None:  # noqa: ARG002
        self.params = {
            "diff_branch_name": self.diff_branch_name,
            "tracking_id": self.tracking_id.serialize() if self.tracking_id else None,
            "diff_id": self.diff_id,
        }
        query = """
MATCH (diff_root:DiffRoot)
WHERE (diff_root.is_merged IS NULL OR diff_root.is_merged <> TRUE)
AND (
    ($diff_id IS NOT NULL AND diff_root.uuid = $diff_id)
    OR ($tracking_id IS NOT NULL AND diff_root.tracking_id = $tracking_id AND diff_root.diff_branch = $diff_branch_name)
)
WITH (
    exists(
        (diff_root)-[:DIFF_HAS_NODE]->(:DiffNode)-[:DIFF_HAS_CONFLICT]->(:DiffConflict)
    ) OR exists(
        (diff_root)-[:DIFF_HAS_NODE]->(:DiffNode)
        -[:DIFF_HAS_ATTRIBUTE]->(:DiffAttribute)
        -[:DIFF_HAS_PROPERTY]->(:DiffProperty)
        -[:DIFF_HAS_CONFLICT]->(:DiffConflict)
    ) OR exists(
        (diff_root)-[:DIFF_HAS_NODE]->(:DiffNode)
        -[:DIFF_HAS_RELATIONSHIP]->(:DiffRelationship)
        -[:DIFF_HAS_ELEMENT]->(:DiffRelationshipElement)
        -[:DIFF_HAS_CONFLICT]->(:DiffConflict)
    ) OR exists(
        (diff_root)-[:DIFF_HAS_NODE]->(:DiffNode)
        -[:DIFF_HAS_RELATIONSHIP]->(:DiffRelationship)
        -[:DIFF_HAS_ELEMENT]->(:DiffRelationshipElement)
        -[:DIFF_HAS_PROPERTY]->(:DiffProperty)
        -[:DIFF_HAS_CONFLICT]->(:DiffConflict)
    )
) AS has_conflict
RETURN has_conflict
        """
        self.return_labels = ["has_conflict"]
        self.add_to_query(query=query)

    async def has_conflict(self) -> bool:
        result = self.get_result()
        return bool(result and result.get("has_conflict"))



from typing import Any

from graphene import BigInt, Boolean, Field, InputObjectType, Int, String
from graphene.types.generic import GenericScalar

from infrahub.core import registry
from infrahub.graphql.types.attribute import GenericPoolInput


class BaseAttributeCreate(InputObjectType):
    is_visible = Boolean(required=False)
    is_protected = Boolean(required=False)
    source = String(required=False)
    owner = String(required=False)

    @classmethod
    def __init_subclass__(cls, **kwargs: dict[str, Any]) -> None:
        super().__init_subclass__(**kwargs)
        registry.input_type[cls.__name__] = cls


class BaseAttributeUpdate(InputObjectType):
    is_default = Boolean(required=False)
    is_visible = Boolean(required=False)
    is_protected = Boolean(required=False)
    source = String(required=False)
    owner = String(required=False)

    @classmethod
    def __init_subclass__(cls, **kwargs: dict[str, Any]) -> None:
        super().__init_subclass__(**kwargs)
        registry.input_type[cls.__name__] = cls


class TextAttributeCreate(BaseAttributeCreate):
    value = String(required=False)


class TextAttributeUpdate(BaseAttributeUpdate):
    value = String(required=False)


class StringAttributeCreate(BaseAttributeCreate):
    value = String(required=False)


class StringAttributeUpdate(BaseAttributeUpdate):
    value = String(required=False)


class NumberAttributeCreate(BaseAttributeCreate):
    value = BigInt(required=False)
    from_pool = Field(GenericPoolInput, required=False)


class NumberAttributeUpdate(BaseAttributeUpdate):
    value = BigInt(required=False)
    from_pool = Field(GenericPoolInput, required=False)


class IntAttributeCreate(BaseAttributeCreate):
    value = Int(required=False)


class IntAttributeUpdate(BaseAttributeUpdate):
    value = Int(required=False)


class CheckboxAttributeCreate(BaseAttributeCreate):
    value = Boolean(required=False)


class CheckboxAttributeUpdate(BaseAttributeUpdate):
    value = Boolean(required=False)


class BoolAttributeCreate(BaseAttributeCreate):
    value = Boolean(required=False)


class BoolAttributeUpdate(BaseAttributeUpdate):
    value = Boolean(required=False)


class ListAttributeCreate(BaseAttributeCreate):
    value = GenericScalar(required=False)


class ListAttributeUpdate(BaseAttributeUpdate):
    value = GenericScalar(required=False)


class JSONAttributeCreate(BaseAttributeCreate):
    value = GenericScalar(required=False)


class JSONAttributeUpdate(BaseAttributeUpdate):
    value = GenericScalar(required=False)


class AnyAttributeCreate(BaseAttributeCreate):
    value = GenericScalar(required=False)


class AnyAttributeUpdate(BaseAttributeUpdate):
    value = GenericScalar(required=False)
